# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


import os
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version
from pathlib import Path
from shutil import copyfile

DOCS_DIR = Path(os.path.dirname(__file__))
ROOT_DIR = (DOCS_DIR / "..").resolve()

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "reuse"
copyright = "2017, Free Software Foundation Europe. CC-BY-SA-4.0"
author = "Free Software Foundation Europe"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
try:
    # The full version, including alpha/beta/rc tags.
    release = get_version("reuse")
except PackageNotFoundError:
    release = "6.1.0"

# The short X.Y.Z version.
version = ".".join(release.split(".")[:3])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinxcontrib.apidoc",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Extensions configuration ------------------------------------------------

apidoc_module_dir = str(ROOT_DIR / "src/reuse")
# apidoc_output_dir = "api"
apidoc_excluded_paths = ["cli"]
apidoc_separate_modules = True
apidoc_toc_file = False
apidoc_extra_args = ["--maxdepth", "2"]

autodoc_member_order = "bysource"

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_logo = "reuse-r-only.svg"
html_static_path = ["_static"]

# -- Options for man output --------------------------------------------------
man_pages = [
    (
        "man/reuse",
        "reuse",
        "A tool for compliance with the REUSE recommendations",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-annotate",
        "reuse-annotate",
        "Add REUSE information to files",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-convert-dep5",
        "reuse-convert-dep5",
        "Convert .reuse/dep5 to REUSE.toml",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-download",
        "reuse-download",
        "Download license files",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-lint",
        "reuse-lint",
        "Verify whether a project is compliant with the REUSE Specification",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-lint-file",
        "reuse-lint-file",
        "Verify whether the specified files are compliant with the REUSE"
        " Specification",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-spdx",
        "reuse-spdx",
        "Generate SPDX bill of materials",
        "Free Software Foundation Europe",
        1,
    ),
    (
        "man/reuse-supported-licenses",
        "reuse-supported-licenses",
        "Print a list of supported licenses",
        "Free Software Foundation Europe",
        1,
    ),
]
manpages_url = (
    "https://reuse.readthedocs.io/en/v{version}/man/{page}.html".format(
        version=version, page="{page}"
    )
)

# -- Custom ------------------------------------------------------------------


def copy_markdown(_):
    """Copy the markdown files from the root of the project into the docs/
    directory.
    """
    copyfile(ROOT_DIR / "README.md", DOCS_DIR / "readme.md")
    copyfile(ROOT_DIR / "CHANGELOG.md", DOCS_DIR / "history.md")
    # this used to be renamed to 'contributing.md', but this caused a conflict
    # with the ToC in the README.
    copyfile(ROOT_DIR / "CONTRIBUTING.md", DOCS_DIR / "contribute.md")


def setup(app):
    app.connect("builder-inited", copy_markdown)



#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Script called by poetry. The API used by poetry is unstable, but let's hope
this stays functional.
"""

import glob
import logging
import os
import shutil
import subprocess
from pathlib import Path

_LOGGER = logging.getLogger(__name__)
ROOT_DIR = Path(os.path.dirname(__file__))
BUILD_DIR = ROOT_DIR / "build"
PO_DIR = ROOT_DIR / "po"


def mkdir_p(path):
    """Make directory and its parents."""
    Path(path).mkdir(parents=True, exist_ok=True)


def rm_fr(path):
    """Force-remove directory."""
    path = Path(path)
    if path.exists():
        shutil.rmtree(path)


def main():
    """Compile .mo files and move them into src directory."""
    rm_fr(BUILD_DIR)
    mkdir_p(BUILD_DIR)

    msgfmt = None
    for executable in ["msgfmt", "msgfmt.py", "msgfmt3.py"]:
        msgfmt = shutil.which(executable)
        if msgfmt:
            break

    if msgfmt:
        po_files = glob.glob(f"{PO_DIR}/*.po")
        mo_files = []

        # Compile
        for po_file in po_files:
            _LOGGER.info(f"compiling {po_file}")
            lang_dir = (
                BUILD_DIR / "reuse/locale" / Path(po_file).stem / "LC_MESSAGES"
            )
            mkdir_p(lang_dir)
            destination = Path(lang_dir) / "reuse.mo"
            subprocess.run(
                [
                    msgfmt,
                    "-o",
                    str(destination),
                    str(po_file),
                ],
                check=True,
            )
            mo_files.append(destination)

        # Move compiled files into src
        rm_fr(ROOT_DIR / "src/reuse/locale")
        for mo_file in mo_files:
            relative = (
                ROOT_DIR / Path("src") / os.path.relpath(mo_file, BUILD_DIR)
            )
            _LOGGER.info(f"copying {mo_file} to {relative}")
            mkdir_p(relative.parent)
            shutil.copyfile(mo_file, relative)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for reuse.report"""


import os
import re
import warnings
from inspect import cleandoc
from textwrap import dedent

from conftest import cpython, posix

from reuse.copyright import SourceType
from reuse.project import Project
from reuse.report import FileReport, ProjectReport, ProjectSubsetReport

# REUSE-IgnoreStart


class TestGenerateFileReport:
    """Tests for FileReport.generate."""

    def test_file_simple(self, fake_repository, add_license_concluded):
        """An extremely simple generate test, just to see if the function
        doesn't crash.
        """
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project,
            "src/source_code.py",
            add_license_concluded=add_license_concluded,
        )

        assert result.licenses_in_file == ["GPL-3.0-or-later"]
        assert (
            result.license_concluded == "GPL-3.0-or-later"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert result.copyright == "SPDX-FileCopyrightText: 2017 Jane Doe"
        assert not result.missing_licenses

    def test_file_from_different_cwd(
        self, fake_repository, add_license_concluded
    ):
        """Another simple generate test, but from a different CWD."""
        os.chdir("/")
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project,
            fake_repository / "src/source_code.py",
            add_license_concluded=add_license_concluded,
        )
        assert result.licenses_in_file == ["GPL-3.0-or-later"]
        assert (
            result.license_concluded == "GPL-3.0-or-later"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert result.copyright == "SPDX-FileCopyrightText: 2017 Jane Doe"
        assert not result.missing_licenses

    def test_file_missing_license(self, fake_repository, add_license_concluded):
        """Simple generate test with a missing license."""
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: BSD-3-Clause"
        )
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project, "foo.py", add_license_concluded=add_license_concluded
        )

        assert result.copyright == ""
        assert result.licenses_in_file == ["BSD-3-Clause"]
        assert (
            result.license_concluded == "BSD-3-Clause"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert result.missing_licenses == {"BSD-3-Clause"}

    def test_file_bad_license(self, fake_repository, add_license_concluded):
        """Simple generate test with a bad license."""
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: fakelicense"
        )
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project, "foo.py", add_license_concluded=add_license_concluded
        )

        assert result.copyright == ""
        assert result.licenses_in_file == ["fakelicense"]
        assert (
            result.license_concluded == "fakelicense"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert result.missing_licenses == {"fakelicense"}

    def test_license_contains_plus(
        self, fake_repository, add_license_concluded
    ):
        """Given a license expression akin to Apache-1.0+,
        LICENSES/Apache-1.0.txt should be an appropriate license file.
        """
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: Apache-1.0+"
        )
        (fake_repository / "LICENSES/Apache-1.0.txt").touch()
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project, "foo.py", add_license_concluded=add_license_concluded
        )

        assert result.copyright == ""
        assert result.licenses_in_file == ["Apache-1.0+"]
        assert (
            result.license_concluded == "Apache-1.0+"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert not result.missing_licenses

    def test_sorted_copyright(self, empty_directory):
        """The copyright notices are sorted like CopyrightNotice, not as a
        string.
        """
        (empty_directory / "foo.py").write_text(
            cleandoc(
                """
                SPDX-FileCopyrightText: 2019 Alice
                Copyright Bob
                © 2017 Jane Doe
                """
            ),
            encoding="utf-8",
        )
        project = Project.from_directory(empty_directory)
        result = FileReport.generate(project, "foo.py")

        assert result.copyright == cleandoc(
            """
            © 2017 Jane Doe
            SPDX-FileCopyrightText: 2019 Alice
            Copyright Bob
            """
        )

    def test_exception(self, fake_repository, add_license_concluded):
        """Simple generate test to test if the exception is detected."""
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project,
            "src/exception.py",
            add_license_concluded=add_license_concluded,
        )

        assert set(result.licenses_in_file) == {
            "GPL-3.0-or-later",
            "Autoconf-exception-3.0",
        }
        assert (
            result.license_concluded
            == "GPL-3.0-or-later WITH Autoconf-exception-3.0"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert result.copyright == "SPDX-FileCopyrightText: 2017 Jane Doe"
        assert not result.missing_licenses

    def test_no_licenses(self, fake_repository, add_license_concluded):
        """Test behavior when no license information is present in the file"""
        (fake_repository / "foo.py").write_text("")
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project, "foo.py", add_license_concluded=add_license_concluded
        )

        assert result.copyright == ""
        assert not result.licenses_in_file
        assert (
            result.license_concluded == "NONE"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert not result.missing_licenses

    def test_multiple_licenses(self, fake_repository, add_license_concluded):
        """Test that all licenses are included in LicenseConcluded"""
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project,
            "src/multiple_licenses.rs",
            add_license_concluded=add_license_concluded,
        )

        assert result.copyright == "SPDX-FileCopyrightText: 2022 Jane Doe"
        assert set(result.licenses_in_file) == {
            "GPL-3.0-or-later",
            "Apache-2.0",
            "CC0-1.0",
            "Autoconf-exception-3.0",
        }
        assert (
            result.license_concluded
            == "GPL-3.0-or-later AND (Apache-2.0 OR CC0-1.0"
            " WITH Autoconf-exception-3.0)"
            if add_license_concluded
            else "NOASSERTION"
        )
        assert not result.missing_licenses

    def test_invalid_spdx_expression_add_license_concluded(
        self, fake_repository, add_license_concluded
    ):
        """LicenseConcluded is NOASSERTION if there is an invalid SPDX License
        Expression.
        """
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: MIT OR"
        )
        project = Project.from_directory(fake_repository)
        result = FileReport.generate(
            project,
            "foo.py",
            add_license_concluded=add_license_concluded,
        )
        assert result.licenses_in_file == ["MIT OR"]
        assert result.license_concluded == "NOASSERTION"

    def test_to_dict_lint_source_information(
        self,
        fake_repository_dep5,
    ):
        """When a file is covered both by DEP5 and its file header, the lint
        dict should correctly convey the source information.
        """
        (fake_repository_dep5 / "doc/foo.py").write_text(
            dedent(
                """
                SPDX-License-Identifier: MIT OR 0BSD
                SPDX-FileCopyrightText: in file"""
            )
        )
        project = Project.from_directory(fake_repository_dep5)
        with warnings.catch_warnings(record=True):
            report = FileReport.generate(
                project,
                "doc/foo.py",
            )
            result = report.to_dict_lint()
        assert result["path"] == "doc/foo.py"
        assert len(result["copyrights"]) == 2
        assert (
            result["copyrights"][0]["source_type"]
            != result["copyrights"][1]["source_type"]
        )
        for copyright_ in result["copyrights"]:
            if copyright_["source_type"] == SourceType.DEP5.value:
                assert copyright_["source"] == ".reuse/dep5"
                assert (
                    copyright_["value"]
                    == "SPDX-FileCopyrightText: 2017 Jane Doe"
                )
            elif copyright_["source_type"] == SourceType.FILE_HEADER.value:
                assert copyright_["source"] == "doc/foo.py"
                assert copyright_["value"] == "SPDX-FileCopyrightText: in file"

        assert len(result["spdx_expressions"]) == 2
        assert (
            result["spdx_expressions"][0]["source_type"]
            != result["spdx_expressions"][1]["source_type"]
        )
        for expression in result["spdx_expressions"]:
            if expression["source_type"] == SourceType.DEP5.value:
                assert expression["source"] == ".reuse/dep5"
                assert expression["value"] == "CC0-1.0"
            elif expression["source_type"] == SourceType.FILE_HEADER.value:
                assert expression["source"] == "doc/foo.py"
                assert expression["value"] == "MIT OR 0BSD"
            assert expression["is_valid"]

    def test_dict_to_lint_invalid_spdx_expression(self, fake_repository):
        """If an SPDX License Expression is invalid, dict_to_lint conveys this
        information.
        """
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: MIT OR"
        )
        project = Project.from_directory(fake_repository)
        report = FileReport.generate(
            project,
            "foo.py",
        )
        result = report.to_dict_lint()
        assert result["path"] == "foo.py"
        assert len(result["copyrights"]) == 0
        assert len(result["spdx_expressions"]) == 1
        expression = result["spdx_expressions"][0]
        assert expression["value"] == "MIT OR"
        assert not expression["is_valid"]


class TestGenerateProjectReport:
    """Tests for ProjectReport.generate."""

    def test_simple(self, fake_repository, multiprocessing):
        """Simple generate test, just to see if it sort of works."""
        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert not result.bad_licenses
        assert not result.licenses_without_extension
        assert not result.missing_licenses
        assert not result.unused_licenses
        assert result.used_licenses
        assert not result.read_errors
        assert result.file_reports

    def test_licenses_without_extension(self, fake_repository, multiprocessing):
        """Licenses without extension are detected."""
        (fake_repository / "LICENSES/CC0-1.0.txt").rename(
            fake_repository / "LICENSES/CC0-1.0"
        )

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert "CC0-1.0" in result.licenses_without_extension

    def test_missing_license(self, fake_repository, multiprocessing):
        """Missing licenses are detected."""
        (fake_repository / "LICENSES/GPL-3.0-or-later.txt").unlink()

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert "GPL-3.0-or-later" in result.missing_licenses
        assert not result.bad_licenses

    def test_bad_license(self, fake_repository, multiprocessing):
        """Bad licenses are detected."""
        (fake_repository / "LICENSES/bad.txt").write_text("foo")

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert result.bad_licenses
        assert not result.missing_licenses

    def test_unused_license(self, fake_repository, multiprocessing):
        """Unused licenses are detected."""
        (fake_repository / "LICENSES/MIT.txt").write_text("foo")

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert result.unused_licenses == {"MIT"}

    def test_unused_license_plus(self, fake_repository, multiprocessing):
        """Apache-1.0+ is not an unused license if LICENSES/Apache-1.0.txt
        exists.

        Furthermore, Apache-1.0+ is separately identified as a used license.
        """
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: Apache-1.0+"
        )
        (fake_repository / "bar.py").write_text(
            "SPDX-License-Identifier: Apache-1.0"
        )
        (fake_repository / "LICENSES/Apache-1.0.txt").touch()

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert not result.unused_licenses
        assert {"Apache-1.0", "Apache-1.0+"}.issubset(result.used_licenses)

    def test_unused_license_plus_only_plus(
        self, fake_repository, multiprocessing
    ):
        """If Apache-1.0+ is the only declared license in the project,
        LICENSES/Apache-1.0.txt should not be an unused license.
        """
        (fake_repository / "foo.py").write_text(
            "SPDX-License-Identifier: Apache-1.0+"
        )
        (fake_repository / "LICENSES/Apache-1.0.txt").touch()

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert not result.unused_licenses
        assert "Apache-1.0+" in result.used_licenses
        assert "Apache-1.0" not in result.used_licenses

    def test_bad_license_can_also_be_missing(
        self, fake_repository, multiprocessing
    ):
        """Bad licenses can also be missing licenses."""
        (fake_repository / "foo.py").write_text("SPDX-License-Identifier: bad")

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert "bad" in result.missing_licenses

    def test_deprecated_license(self, fake_repository, multiprocessing):
        """Deprecated licenses are detected."""
        (fake_repository / "LICENSES/GPL-3.0-or-later.txt").rename(
            fake_repository / "LICENSES/GPL-3.0.txt"
        )

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        assert "GPL-3.0" in result.deprecated_licenses

    @cpython
    @posix
    def test_read_error(self, fake_repository, multiprocessing):
        """Files that cannot be read are added to the read error list."""
        (fake_repository / "bad").write_text("foo")
        (fake_repository / "bad").chmod(0o000)

        project = Project.from_directory(fake_repository)
        result = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )

        # pylint: disable=superfluous-parens
        assert (fake_repository / "bad") in result.read_errors

    def test_to_dict_lint(self, fake_repository, multiprocessing):
        """Generate dictionary output and verify correct ordering."""
        project = Project.from_directory(fake_repository)
        report = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )
        result = report.to_dict_lint()

        # Check if the top three keys are at the beginning of the dictionary
        assert list(result.keys())[:3] == [
            "lint_version",
            "reuse_spec_version",
            "reuse_tool_version",
        ]

        # Check if the recommendation key is at the bottom of the dictionary
        assert list(result.keys())[-1] == "recommendations"

        # Check if the rest of the keys are sorted alphabetically
        assert list(result.keys())[3:-1] == sorted(list(result.keys())[3:-1])

    def test_partial_info_in_toml(self, empty_directory, multiprocessing):
        """Some information is in REUSE.toml, and some is inside of the file."""
        (empty_directory / "REUSE.toml").write_text(
            cleandoc(
                """
                version = 1

                [[annotations]]
                path = "foo.py"
                precedence = "closest"
                SPDX-FileCopyrightText = "Jane Doe"
                # This is ignored because it's in the file!
                SPDX-License-Identifier = "MIT"
                """
            )
        )
        (empty_directory / "foo.py").write_text(
            "# SPDX-License-Identifier: 0BSD"
        )
        project = Project.from_directory(empty_directory)
        report = ProjectReport.generate(
            project, multiprocessing=multiprocessing
        )
        file_report = next(
            report
            for report in report.file_reports
            if report.path.name == "foo.py"
        )
        infos = file_report.reuse_infos
        assert len(infos) == 2
        assert file_report.copyright == "SPDX-FileCopyrightText: Jane Doe"
        assert file_report.licenses_in_file == ["0BSD"]


class TestProjectSubsetReport:
    """Tests for ProjectSubsetReport."""

    def test_simple(self, fake_repository, multiprocessing):
        """Simple generate test."""
        project = Project.from_directory(fake_repository)
        result = ProjectSubsetReport.generate(
            project,
            {fake_repository / "src/custom.py"},
            multiprocessing=multiprocessing,
        )

        assert not result.missing_licenses
        assert not result.read_errors
        assert not result.files_without_licenses
        assert not result.files_without_copyright
        assert len(result.file_reports) == 1

    @cpython
    @posix
    def test_read_error(self, fake_repository, multiprocessing):
        """Files that cannot be read are added to the read error list."""
        (fake_repository / "bad").write_text("foo")
        (fake_repository / "bad").chmod(0o000)

        project = Project.from_directory(fake_repository)
        result = ProjectSubsetReport.generate(
            project, {fake_repository / "bad"}, multiprocessing=multiprocessing
        )

        # pylint: disable=superfluous-parens
        assert (fake_repository / "bad") in result.read_errors

    def test_missing_license(self, fake_repository, multiprocessing):
        """Missing licenses are detected."""
        (fake_repository / "LICENSES/GPL-3.0-or-later.txt").unlink()

        project = Project.from_directory(fake_repository)
        result = ProjectSubsetReport.generate(
            project,
            {fake_repository / "src/exception.py"},
            multiprocessing=multiprocessing,
        )

        assert result.missing_licenses == {
            "GPL-3.0-or-later": {fake_repository / "src/exception.py"}
        }

    def test_missing_copyright_license(self, empty_directory, multiprocessing):
        """Missing copyright and license is detected."""
        (empty_directory / "foo.py").write_text("foo")
        project = Project.from_directory(empty_directory)
        result = ProjectSubsetReport.generate(
            project,
            {empty_directory / "foo.py"},
            multiprocessing=multiprocessing,
        )

        # pylint: disable=superfluous-parens
        assert (empty_directory / "foo.py") in result.files_without_copyright
        assert (empty_directory / "foo.py") in result.files_without_licenses


def test_bill_of_materials(fake_repository, multiprocessing):
    """Generate a bill of materials."""
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project, multiprocessing=multiprocessing)
    # TODO: Actually do something
    bom = report.bill_of_materials()
    created_re = re.compile(
        r"^Created: \d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$", re.MULTILINE
    )
    assert created_re.search(bom) is not None


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for download."""

# pylint: disable=redefined-outer-name,unused-argument,unspecified-encoding

import errno
import os
import shutil
from pathlib import Path
from unittest.mock import create_autospec
from urllib.error import URLError

import pytest
from click.testing import CliRunner

from reuse.cli import download
from reuse.cli.main import main

# REUSE-IgnoreStart


@pytest.fixture()
def mock_put_license_in_file(monkeypatch):
    """Create a mocked version of put_license_in_file."""
    result = create_autospec(download.put_license_in_file)
    monkeypatch.setattr(download, "put_license_in_file", result)
    return result


class TestDownload:
    """Tests for download."""

    def test_simple(self, empty_directory, mock_put_license_in_file):
        """Straightforward test."""
        result = CliRunner().invoke(main, ["download", "0BSD"])

        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "0BSD", Path(os.path.realpath("LICENSES/0BSD.txt")), source=None
        )

    def test_strip_plus(self, empty_directory, mock_put_license_in_file):
        """If downloading LIC+, download LIC instead."""
        result = CliRunner().invoke(main, ["download", "EUPL-1.2+"])

        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "EUPL-1.2",
            Path(os.path.realpath("LICENSES/EUPL-1.2.txt")),
            source=None,
        )

    def test_all(self, fake_repository, mock_put_license_in_file):
        """--all downloads all detected licenses."""
        shutil.rmtree("LICENSES")
        result = CliRunner().invoke(main, ["download", "--all"])

        assert result.exit_code == 0
        for lic in [
            "GPL-3.0-or-later",
            "LicenseRef-custom",
            "Autoconf-exception-3.0",
            "Apache-2.0",
            "CC0-1.0",
        ]:
            mock_put_license_in_file.assert_any_call(
                lic, Path(os.path.realpath(f"LICENSES/{lic}.txt")), source=None
            )

    def test_all_with_plus(self, fake_repository, mock_put_license_in_file):
        """--all downloads EUPL-1.2 if EUPL-1.2+ is detected."""
        Path("foo.py").write_text("# SPDX-License-Identifier: EUPL-1.2+")
        result = CliRunner().invoke(main, ["download", "--all"])

        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_once_with(
            "EUPL-1.2",
            Path(os.path.realpath("LICENSES/EUPL-1.2.txt")),
            source=None,
        )

    def test_all_with_plus_and_non_plus(
        self, fake_repository, mock_put_license_in_file
    ):
        """If both EUPL-1.2 and EUPL-1.2+ is detected, download EUPL-1.2 only
        once.
        """
        Path("foo.py").write_text(
            """
            # SPDX-License-Identifier: EUPL-1.2+
            # SPDX-License-Identifier: EUPL-1.2
            """
        )
        result = CliRunner().invoke(main, ["download", "--all"])

        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_once_with(
            "EUPL-1.2",
            Path(os.path.realpath("LICENSES/EUPL-1.2.txt")),
            source=None,
        )

    def test_all_and_license_mutually_exclusive(self, empty_directory):
        """--all and license args are mutually exclusive."""
        result = CliRunner().invoke(main, ["download", "--all", "0BSD"])
        assert result.exit_code != 0
        assert "are mutually exclusive" in result.output

    def test_all_and_output_mutually_exclusive(self, empty_directory):
        """--all and --output are mutually exclusive."""
        result = CliRunner().invoke(
            main, ["download", "--all", "--output", "foo"]
        )
        assert result.exit_code != 0
        assert "is mutually exclusive with" in result.output

    def test_file_exists(self, fake_repository, mock_put_license_in_file):
        """The to-be-downloaded file already exists."""
        mock_put_license_in_file.side_effect = FileExistsError(
            errno.EEXIST, "", "GPL-3.0-or-later.txt"
        )

        result = CliRunner().invoke(main, ["download", "GPL-3.0-or-later"])

        assert result.exit_code == 1
        assert "GPL-3.0-or-later.txt already exists" in result.output

    def test_exception(self, empty_directory, mock_put_license_in_file):
        """There was an error while downloading the license file."""
        mock_put_license_in_file.side_effect = URLError("test")

        result = CliRunner().invoke(main, ["download", "0BSD"])

        assert result.exit_code == 1
        assert "internet" in result.output

    def test_invalid_spdx(self, empty_directory, mock_put_license_in_file):
        """An invalid SPDX identifier was provided."""
        mock_put_license_in_file.side_effect = URLError("test")

        result = CliRunner().invoke(main, ["download", "does-not-exist"])

        assert result.exit_code == 1
        assert "not a valid SPDX License Identifier" in result.output

    def test_custom_output(self, empty_directory, mock_put_license_in_file):
        """Download the license into a custom file."""
        result = CliRunner().invoke(main, ["download", "-o", "foo", "0BSD"])

        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "0BSD", destination=Path("foo"), source=None
        )

    def test_custom_output_too_many(
        self, empty_directory, mock_put_license_in_file
    ):
        """Providing more than one license with a custom output results in an
        error.
        """
        result = CliRunner().invoke(
            main,
            ["download", "-o", "foo", "0BSD", "GPL-3.0-or-later"],
        )

        assert result.exit_code != 0
        assert (
            "Cannot use '--output' with more than one license" in result.output
        )

    def test_inside_licenses_dir(
        self, fake_repository, mock_put_license_in_file
    ):
        """While inside the LICENSES/ directory, don't create another LICENSES/
        directory.
        """
        os.chdir(fake_repository / "LICENSES")
        result = CliRunner().invoke(main, ["download", "0BSD"])
        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "0BSD", destination=Path("0BSD.txt").absolute(), source=None
        )

    def test_inside_licenses_dir_in_git(
        self, git_repository, mock_put_license_in_file
    ):
        """While inside a random LICENSES/ directory in a Git repository, use
        the root LICENSES/ directory.
        """
        (git_repository / "doc/LICENSES").mkdir()
        os.chdir(git_repository / "doc/LICENSES")
        result = CliRunner().invoke(main, ["download", "0BSD"])
        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "0BSD", destination=Path("../../LICENSES/0BSD.txt"), source=None
        )

    def test_different_root(self, fake_repository, mock_put_license_in_file):
        """Download using a different root."""
        (fake_repository / "new_root").mkdir()

        result = CliRunner().invoke(
            main,
            [
                "--root",
                str((fake_repository / "new_root").resolve()),
                "download",
                "MIT",
            ],
        )
        assert result.exit_code == 0
        mock_put_license_in_file.assert_called_with(
            "MIT", Path("new_root/LICENSES/MIT.txt").resolve(), source=None
        )

    def test_licenseref_no_source(self, empty_directory):
        """Downloading a LicenseRef license creates an empty file."""
        CliRunner().invoke(main, ["download", "LicenseRef-hello"])
        assert (
            empty_directory / "LICENSES/LicenseRef-hello.txt"
        ).read_text() == ""

    def test_licenseref_source_file(
        self,
        empty_directory,
    ):
        """Downloading a LicenseRef license with a source file copies that
        file's contents.
        """
        (empty_directory / "foo.txt").write_text("foo")
        CliRunner().invoke(
            main,
            ["download", "--source", "foo.txt", "LicenseRef-hello"],
        )
        assert (
            empty_directory / "LICENSES/LicenseRef-hello.txt"
        ).read_text() == "foo"

    def test_licenseref_source_dir(self, empty_directory):
        """Downloading a LicenseRef license with a source dir copies the text
        from the corresponding file in the directory.
        """
        (empty_directory / "lics").mkdir()
        (empty_directory / "lics/LicenseRef-hello.txt").write_text("foo")

        CliRunner().invoke(
            main, ["download", "--source", "lics", "LicenseRef-hello"]
        )
        assert (
            empty_directory / "LICENSES/LicenseRef-hello.txt"
        ).read_text() == "foo"

    def test_licenseref_false_source_dir(self, empty_directory):
        """Downloading a LicenseRef license with a source that does not contain
        the license results in an error.
        """
        (empty_directory / "lics").mkdir()

        result = CliRunner().invoke(
            main, ["download", "--source", "lics", "LicenseRef-hello"]
        )
        assert result.exit_code == 1
        assert (
            f"{Path('lics') / 'LicenseRef-hello.txt'} does not exist"
            in result.output
        )


class TestSimilarIdentifiers:
    """Test a private function _similar_spdx_identifiers."""

    # pylint: disable=protected-access

    def test_typo(self):
        """Given a misspelt SPDX License Identifier, suggest a better one."""
        result = download._similar_spdx_identifiers("GPL-3.0-or-lter")

        assert "GPL-3.0-or-later" in result
        assert "AGPL-3.0-or-later" in result
        assert "LGPL-3.0-or-later" in result

    def test_prefix(self):
        """Given an incomplete SPDX License Identifier, suggest a better one."""
        result = download._similar_spdx_identifiers("CC0")

        assert "CC0-1.0" in result


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 Matthias Riße
# SPDX-FileCopyrightText: 2024 Skyler Grey <sky@a.starrysky.fyi>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global fixtures and configuration."""

# pylint: disable=redefined-outer-name,invalid-name

import concurrent.futures
import contextlib
import datetime
import importlib
import logging
import os
import shutil
import subprocess
import sys
from collections.abc import Generator
from inspect import cleandoc
from io import StringIO
from pathlib import Path
from unittest.mock import create_autospec

import pytest
from jinja2 import Environment

os.environ["LC_ALL"] = "C"
os.environ["LANGUAGE"] = ""

# A trick that tries to import the installed version of reuse. If that doesn't
# work, import from the src directory. If that also doesn't work (for some
# reason), then an ImportError is raised.
try:
    # pylint: disable=unused-import
    import reuse
except ImportError:
    sys.path.append(os.path.join(Path(__file__).parent.parent, "src"))
finally:
    from reuse import extract, report
    from reuse._util import setup_logging
    from reuse.global_licensing import ReuseDep5
    from reuse.vcs import GIT_EXE, HG_EXE, JUJUTSU_EXE, PIJUL_EXE

try:
    _chardet = bool(importlib.import_module("chardet"))
except ImportError:
    _chardet = False

CWD = Path.cwd()

TESTS_DIRECTORY = Path(__file__).parent.resolve()
RESOURCES_DIRECTORY = TESTS_DIRECTORY / "resources"

try:
    import pwd

    is_root = pwd.getpwuid(os.getuid()).pw_name == "root"
    is_posix = True
except ImportError:
    is_root = False
    is_posix = False

cpython = pytest.mark.skipif(
    sys.implementation.name != "cpython", reason="only CPython supported"
)
git = pytest.mark.skipif(not GIT_EXE, reason="requires git")
hg = pytest.mark.skipif(not HG_EXE, reason="requires mercurial")
pijul = pytest.mark.skipif(not PIJUL_EXE, reason="requires pijul")
no_root = pytest.mark.xfail(is_root, reason="fails when user is root")
posix = pytest.mark.skipif(not is_posix, reason="Windows not supported")
chardet = pytest.mark.skipif(not _chardet, reason="chardet is not installed")


# REUSE-IgnoreStart


def pytest_addoption(parser):
    """Allows specification of additional commandline options to parse"""
    parser.addoption("--loglevel", action="store", default="DEBUG")


def pytest_configure(config):
    """Called after command line options have been parsed and all plugins and
    initial conftest files been loaded.
    """
    loglevel = config.getoption("loglevel")
    setup_logging(level=logging.getLevelName(loglevel))

    # Disable parallelisation during tests.
    report.ENABLE_PARALLEL = False


def pytest_runtest_setup(item):
    """Called before running a test."""
    # pylint: disable=unused-argument
    # TODO: Awful workaround. In `main`, this environment variable is set under
    # certain conditions. This means that all tests that run _after_ that
    # condition is met also have the environment variable set, because the
    # environment had been changed. There should be a better way to scope this.
    with contextlib.suppress(KeyError):
        del os.environ["_SUPPRESS_DEP5_WARNING"]


def pytest_runtest_teardown(item):
    """Called after running a test."""
    # pylint: disable=unused-argument
    # Make sure to restore CWD
    os.chdir(CWD)


@pytest.fixture(scope="session")
def git_exe() -> str:
    """Run the test with git."""
    if not GIT_EXE:
        pytest.skip("cannot run this test without git")
    return str(GIT_EXE)


@pytest.fixture(params=[True, False])
def optional_git_exe(request, monkeypatch) -> Generator[str | None, None, None]:
    """Run the test with or without git."""
    exe = GIT_EXE if request.param else ""
    monkeypatch.setattr("reuse.vcs.GIT_EXE", exe)
    yield exe


@pytest.fixture(scope="session")
def hg_exe() -> str:
    """Run the test with mercurial (hg)."""
    if not HG_EXE:
        pytest.skip("cannot run this test without mercurial")
    return str(HG_EXE)


@pytest.fixture(params=[True, False])
def optional_hg_exe(request, monkeypatch) -> Generator[str | None, None, None]:
    """Run the test with or without mercurial."""
    exe = HG_EXE if request.param else ""
    monkeypatch.setattr("reuse.vcs.HG_EXE", exe)
    yield exe


@pytest.fixture(scope="session")
def jujutsu_exe() -> str:
    """Run the test with Jujutsu."""
    if not JUJUTSU_EXE:
        pytest.skip("cannot run this test without jujutsu")
    return str(JUJUTSU_EXE)


@pytest.fixture(params=[True, False])
def optional_jujutsu_exe(
    request, monkeypatch
) -> Generator[str | None, None, None]:
    """Run the test with or without Jujutsu."""
    exe = JUJUTSU_EXE if request.param else ""
    monkeypatch.setattr("reuse.vcs.JUJUTSU_EXE", exe)
    yield exe


@pytest.fixture(scope="session")
def pijul_exe() -> str:
    """Run the test with Pijul."""
    if not PIJUL_EXE:
        pytest.skip("cannot run this test without pijul")
    return str(PIJUL_EXE)


@pytest.fixture(params=[True, False])
def optional_pijul_exe(
    request, monkeypatch
) -> Generator[str | None, None, None]:
    """Run the test with or without Pijul."""
    exe = PIJUL_EXE if request.param else ""
    monkeypatch.setattr("reuse.vcs.PIJUL_EXE", exe)
    yield exe


@pytest.fixture(params=[True, False])
def multiprocessing(request, monkeypatch) -> Generator[bool, None, None]:
    """Run the test with or without multiprocessing."""
    monkeypatch.setattr("reuse.report.ENABLE_PARALLEL", True)
    if not request.param:
        monkeypatch.delattr(concurrent.futures, "ProcessPoolExecutor")
    yield request.param


@pytest.fixture(params=["magic", "charset_normalizer", "chardet"])
def encoding_module(request, monkeypatch) -> Generator[bool, None, None]:
    """Run the test with or without libmagic."""
    try:
        if request.param == "magic" and not is_posix:
            pytest.skip("Windows not supported")
        module = importlib.import_module(request.param)
        monkeypatch.setattr("reuse.extract._ENCODING_MODULE", module)
        yield request.param
    except ImportError:
        pytest.skip(f"'{request.param}' could not be imported")


@pytest.fixture(params=[True, False])
def add_license_concluded(request) -> Generator[bool, None, None]:
    yield request


@pytest.fixture()
def empty_directory(tmpdir_factory) -> Path:
    """Create a temporary empty directory."""
    directory = Path(str(tmpdir_factory.mktemp("empty_directory")))

    os.chdir(str(directory))
    return directory


@pytest.fixture(scope="session")
def _cached_fake_repository(tmp_path_factory) -> Path:
    """Create a temporary fake repository."""
    directory = tmp_path_factory.mktemp("fake_repository")
    shutil.copytree(
        RESOURCES_DIRECTORY / "fake_repository", directory, dirs_exist_ok=True
    )

    # Get rid of those pesky pyc files.
    shutil.rmtree(directory / "src/__pycache__", ignore_errors=True)

    return directory


@pytest.fixture()
def fake_repository(_cached_fake_repository, tmp_path) -> Path:
    """Create a temporary fake repository."""
    shutil.copytree(_cached_fake_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture()
def fake_repository_reuse_toml(fake_repository) -> Path:
    """Add REUSE.toml to the fake repo."""
    shutil.copy(
        RESOURCES_DIRECTORY / "REUSE.toml", fake_repository / "REUSE.toml"
    )
    (fake_repository / "doc/index.rst").touch()
    return fake_repository


@pytest.fixture()
def fake_repository_dep5(fake_repository) -> Path:
    """Add .reuse/dep5 to the fake repo."""
    (fake_repository / ".reuse").mkdir(exist_ok=True)
    shutil.copy(RESOURCES_DIRECTORY / "dep5", fake_repository / ".reuse/dep5")
    (fake_repository / "doc/index.rst").touch()
    return fake_repository


def _repo_contents(
    fake_repository, ignore_filename=".gitignore", ignore_prefix=""
):
    """Generate contents for a vcs repository.

    Currently defaults to git-like behavior for ignoring files with
    the expectation that other tools can be configured to ignore files
    by just chanigng the ignore-file-name and enabling git-like behavior
    with a prefix line in the ignore file.
    """
    gitignore = ignore_prefix + (
        "# SPDX-License-Identifier: CC0-1.0\n"
        "# SPDX-FileCopyrightText: 2017 Jane Doe\n"
        "*.pyc\nbuild"
    )
    (fake_repository / ignore_filename).write_text(gitignore)
    (fake_repository / "LICENSES/CC0-1.0.txt").write_text("License text")

    for file_ in (fake_repository / "src").iterdir():
        if file_.suffix == ".py":
            file_.with_suffix(".pyc").write_text("foo")

    build_dir = fake_repository / "build"
    build_dir.mkdir()
    (build_dir / "hello.py").write_text("foo")


@pytest.fixture(scope="session")
def _cached_git_repository(
    _cached_fake_repository: Path, tmp_path_factory, git_exe: str
) -> Path:
    """Create a git repository with ignored files."""
    directory = tmp_path_factory.mktemp("cached_git_directory")
    shutil.copytree(_cached_fake_repository, directory, dirs_exist_ok=True)
    os.chdir(directory)
    _repo_contents(directory)

    subprocess.run([git_exe, "init", str(directory)], check=True)
    Path(".git/config").write_text(
        cleandoc(
            """
            [user]
              name = Example
              email = example@example.com
            [commit]
              gpgSign = false
            """
        ),
        encoding="utf-8",
    )
    subprocess.run([git_exe, "add", str(directory)], check=True)
    subprocess.run(
        [
            git_exe,
            "commit",
            "-m",
            "initial",
        ],
        check=True,
    )

    return directory


@pytest.fixture()
def git_repository(_cached_git_repository, tmp_path) -> Path:
    """Create a git repository with ignored files."""
    shutil.copytree(_cached_git_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def _cached_hg_repository(
    _cached_fake_repository: Path, tmp_path_factory, hg_exe: str
) -> Path:
    """Create a mercurial repository with ignored files."""
    directory = tmp_path_factory.mktemp("cached_hg_repository")
    shutil.copytree(_cached_fake_repository, directory, dirs_exist_ok=True)
    os.chdir(directory)
    _repo_contents(
        directory,
        ignore_filename=".hgignore",
        ignore_prefix="syntax:glob",
    )

    subprocess.run([hg_exe, "init", "."], check=True)
    subprocess.run([hg_exe, "addremove"], check=True)
    subprocess.run(
        [
            hg_exe,
            "commit",
            "--user",
            "Example <example@example.com>",
            "-m",
            "initial",
        ],
        check=True,
    )

    return directory


@pytest.fixture()
def hg_repository(_cached_hg_repository, tmp_path) -> Path:
    """Create a mercurial repository with ignored files."""
    shutil.copytree(_cached_hg_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def _cached_jujutsu_repository(
    _cached_fake_repository: Path, tmp_path_factory, jujutsu_exe: str
) -> Path:
    """Create a jujutsu repository with ignored files."""
    directory = tmp_path_factory.mktemp("cached_jujutsu_repository")
    shutil.copytree(_cached_fake_repository, directory, dirs_exist_ok=True)
    os.chdir(directory)
    _repo_contents(directory)

    subprocess.run([jujutsu_exe, "git", "init", str(directory)], check=True)

    return directory


@pytest.fixture()
def jujutsu_repository(_cached_jujutsu_repository, tmp_path) -> Path:
    """Create a jujutsu repository with ignored files."""
    shutil.copytree(_cached_jujutsu_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def _cached_pijul_repository(
    _cached_fake_repository: Path, tmp_path_factory, pijul_exe: str
) -> Path:
    """Create a pijul repository with ignored files."""
    directory = tmp_path_factory.mktemp("cached_pijul_repository")
    shutil.copytree(_cached_fake_repository, directory, dirs_exist_ok=True)
    os.chdir(directory)
    _repo_contents(
        directory,
        ignore_filename=".ignore",
    )

    subprocess.run([pijul_exe, "init", "."], check=True)
    subprocess.run([pijul_exe, "add", "--recursive", "."], check=True)
    subprocess.run(
        [
            pijul_exe,
            "record",
            "--all",
            "--message",
            "initial",
        ],
        check=True,
    )

    return directory


@pytest.fixture()
def pijul_repository(_cached_pijul_repository, tmp_path) -> Path:
    """Create a pijul repository with ignored files."""
    shutil.copytree(_cached_pijul_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session", params=["submodule-add", "manual"])
def _cached_submodule_repository(
    _cached_git_repository: Path, git_exe: str, tmp_path_factory, request
) -> Path:
    """Create a git repository that contains a submodule."""
    directory = tmp_path_factory.mktemp("cached_submodule_repository")
    shutil.copytree(_cached_git_repository, directory, dirs_exist_ok=True)
    header = cleandoc(
        """
            SPDX-FileCopyrightText: 2019 Jane Doe

            SPDX-License-Identifier: CC0-1.0
            """
    )

    submodule = tmp_path_factory.mktemp("submodule")
    (submodule / "foo.py").write_text(header, encoding="utf-8")

    os.chdir(submodule)

    subprocess.run([git_exe, "init", str(submodule)], check=True)
    Path(".git/config").write_text(
        cleandoc(
            """
            [user]
              name = Example
              email = example@example.com
            [commit]
              gpgSign = false
            """
        ),
        encoding="utf-8",
    )

    subprocess.run([git_exe, "add", str(submodule)], check=True)
    subprocess.run(
        [
            git_exe,
            "commit",
            "-m",
            "initial",
        ],
        check=True,
    )

    os.chdir(directory)

    if request.param == "submodule-add":
        subprocess.run(
            [
                git_exe,
                # https://git-scm.com/docs/git-config#Documentation/git-config.txt-protocolallow
                #
                # This circumvents a bug/behaviour caused by CVE-2022-39253
                # where you cannot use `git submodule add repository path` where
                # repository is a file on the filesystem.
                "-c",
                "protocol.file.allow=always",
                "submodule",
                "add",
                str(submodule.resolve()),
                "submodule",
            ],
            check=True,
        )
    elif request.param == "manual":
        subprocess.run(
            [git_exe, "clone", str(submodule.resolve()), "submodule"],
            check=True,
        )
        with open(
            directory / ".gitmodules", mode="a", encoding="utf-8"
        ) as gitmodules_file:
            gitmodules_file.write(
                cleandoc(
                    f"""
                    [submodule "submodule"]
                      path = submodule
                      url = {submodule.resolve().as_posix()}
                    """
                )
            )
        subprocess.run(
            [
                git_exe,
                "add",
                "--no-warn-embedded-repo",
                ".gitmodules",
                "submodule",
            ],
            check=True,
        )

    subprocess.run(
        [git_exe, "commit", "-m", "add submodule"],
        check=True,
    )
    (directory / ".gitmodules.license").write_text(header)

    return directory


@pytest.fixture()
def submodule_repository(_cached_submodule_repository, tmp_path) -> Path:
    """Create a git repository that contains a submodule."""
    shutil.copytree(_cached_submodule_repository, tmp_path, dirs_exist_ok=True)
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture()
def subproject_repository(fake_repository: Path) -> Path:
    """Add a Meson subproject to the fake repo."""
    (fake_repository / "meson.build").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: 2022 Jane Doe
            SPDX-License-Identifier: CC0-1.0
            """
        )
    )
    subprojects_dir = fake_repository / "subprojects"
    subprojects_dir.mkdir()
    libfoo_dir = subprojects_dir / "libfoo"
    libfoo_dir.mkdir()
    # ./subprojects/foo.wrap has license and linter succeeds
    (subprojects_dir / "foo.wrap").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: 2022 Jane Doe
            SPDX-License-Identifier: CC0-1.0
            """
        )
    )
    # ./subprojects/libfoo/foo.c misses license but is ignored
    (libfoo_dir / "foo.c").write_text("foo")
    return fake_repository


@pytest.fixture(scope="session")
def reuse_dep5():
    """Create a ReuseDep5 object."""
    return ReuseDep5.from_file(RESOURCES_DIRECTORY / "dep5")


@pytest.fixture()
def stringio():
    """Create a StringIO object."""
    return StringIO()


@pytest.fixture()
def binary_string():
    """Create a binary string."""
    return bytes(range(256))


@pytest.fixture()
def template_simple_source():
    """Source code of simple Jinja2 template."""
    return cleandoc(
        """
        Hello, world!

        {% for copyright_line in copyright_lines %}
        {{ copyright_line }}
        {% endfor %}

        {% for expression in spdx_expressions %}
        SPDX-License-Identifier: {{ expression }}
        {% endfor %}
        """
    )


@pytest.fixture()
def template_simple(template_simple_source):
    """Provide a simple Jinja2 template."""
    env = Environment(trim_blocks=True)
    return env.from_string(template_simple_source)


@pytest.fixture()
def template_no_spdx_source():
    """Source code of Jinja2 template without SPDX lines."""
    return "Hello, world"


@pytest.fixture()
def template_no_spdx(template_no_spdx_source):
    """Provide a Jinja2 template without SPDX lines."""
    env = Environment(trim_blocks=True)
    return env.from_string(template_no_spdx_source)


@pytest.fixture()
def template_commented_source():
    """Source code of a simple Jinja2 template that is already commented."""
    return cleandoc(
        """
        # Hello, world!
        #
        {% for copyright_line in copyright_lines %}
        # {{ copyright_line }}
        {% endfor %}
        #
        {% for expression in spdx_expressions %}
        # SPDX-License-Identifier: {{ expression }}
        {% endfor %}
        """.replace(
            "spdx-Lic", "SPDX-Lic"
        )
    )


@pytest.fixture()
def template_commented(template_commented_source):
    """Provide a Jinja2 template that is already commented."""
    env = Environment(trim_blocks=True)
    return env.from_string(template_commented_source)


@pytest.fixture()
def mock_date_today(monkeypatch):
    """Mock away datetime.date.today to always return 2018."""
    date = create_autospec(datetime.date)
    date.today.return_value = datetime.date(2018, 1, 1)
    monkeypatch.setattr(datetime, "date", date)


@pytest.fixture(
    params=[[], ["John Doe"], ["John Doe", "Alice Doe"]],
    ids=["None", "John", "John and Alice"],
)
def contributors(request):
    """Provide contributors for SPDX-FileContributor field generation"""
    yield request.param


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmen@carmenbianca.eu>
# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for reuse.cli.main."""

from click.testing import CliRunner

from reuse import __version__
from reuse.cli.main import main


class TestMain:
    """Collect all tests for main."""

    def test_help_is_default(self):
        """--help is optional."""
        without_help = CliRunner().invoke(main, [])
        with_help = CliRunner().invoke(main, ["--help"])
        assert without_help.output == with_help.output
        assert without_help.exit_code in {0, 2}
        assert with_help.exit_code == 0
        assert with_help.output.startswith("Usage: reuse")

    def test_version(self):
        """--version returns the correct version."""
        result = CliRunner().invoke(main, ["--version"])
        assert result.output.startswith(f"reuse, version {__version__}\n")
        assert "This program is free software:" in result.output
        assert "GNU General Public License" in result.output



# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for convert_dep5."""

# pylint: disable=line-too-long

from io import StringIO

from debian.copyright import Copyright

from reuse._util import cleandoc_nl
from reuse.convert_dep5 import toml_from_dep5


def test_toml_from_dep5_single_file():
    """Correctly convert a DEP5 file with a single file."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: hello.txt
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_asterisks():
    """Single asterisks get converted to double asterisks. Double asterisks get
    left alone.
    """
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: */**/***
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = "**/**/***"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_multiple_files_in_paragraph():
    """Correctly convert a DEP5 file with a more files in a paragraph."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: hello.txt foo*.txt
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = ["hello.txt", "foo**.txt"]
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_multiple_paragraphs():
    """Correctly convert a DEP5 file with multiple paragraphs."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: hello.txt
            Copyright: 2018 Jane Doe
            License: MIT

            Files: world.txt
            Copyright: 2018 John Doe
            License: 0BSD
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"

        [[annotations]]
        path = "world.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 John Doe"
        SPDX-License-Identifier = "0BSD"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_multiple_copyright():
    """Correctly convert a DEP5 file with multiple copyright holders."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: hello.txt
            Copyright: 2018 Jane Doe
                2018 John Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = ["2018 Jane Doe", "2018 John Doe"]
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_comments():
    """Optionally include comments."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

            Files: hello.txt
            Copyright: 2018 Jane Doe
            License: MIT
            Comment: hello

            Files: world.txt
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        SPDX-FileComment = "hello"

        [[annotations]]
        path = "world.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_header():
    """Optionally include header fields."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
            Upstream-Name: Some project
            Upstream-Contact: Jane Doe
            Source: https://example.com/
            Disclaimer: Some rights reserved

            Files: hello.txt
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1
        SPDX-PackageName = "Some project"
        SPDX-PackageSupplier = "Jane Doe"
        SPDX-PackageDownloadLocation = "https://example.com/"
        SPDX-PackageComment = "Some rights reserved"

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_header_multiple_contacts():
    """Return a list of contacts."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
            Upstream-Contact: Jane Doe
                John Doe

            Files: hello.txt
            Copyright: 2018 Jane Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1
        SPDX-PackageSupplier = ["Jane Doe", "John Doe"]

        [[annotations]]
        path = "hello.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected


def test_toml_from_dep5_man_example():
    """Test the example from the man page."""
    text = StringIO(
        cleandoc_nl(
            """
            Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
            Upstream-Name: Some project
            Upstream-Contact: Jane Doe
            Source: https://example.com/
            Disclaimer: Some rights reserved

            Files: hello*.txt
            Copyright: 2018 Jane Doe
            License: MIT
            Comment: hello world

            Files: foo bar
            Copyright: 2018 Jane Doe
                2019 John Doe
            License: MIT
            """
        )
    )
    expected = cleandoc_nl(
        """
        version = 1
        SPDX-PackageName = "Some project"
        SPDX-PackageSupplier = "Jane Doe"
        SPDX-PackageDownloadLocation = "https://example.com/"
        SPDX-PackageComment = "Some rights reserved"

        [[annotations]]
        path = "hello**.txt"
        precedence = "aggregate"
        SPDX-FileCopyrightText = "2018 Jane Doe"
        SPDX-License-Identifier = "MIT"
        SPDX-FileComment = "hello world"

        [[annotations]]
        path = ["foo", "bar"]
        precedence = "aggregate"
        SPDX-FileCopyrightText = ["2018 Jane Doe", "2019 John Doe"]
        SPDX-License-Identifier = "MIT"
        """
    )
    assert toml_from_dep5(Copyright(text)) == expected



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2021 Michael Weimann
# SPDX-FileCopyrightText: 2025 Shun Sakai <sorairolake@protonmail.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for supported-licenses."""

import json
import re

from click.testing import CliRunner

from reuse.cli.main import main


class TestSupportedLicenses:
    """Tests for supported-licenses."""

    def test_simple(self):
        """Invoke the supported-licenses command and check whether the result
        contains at least one license in the expected format.
        """

        result = CliRunner().invoke(main, ["supported-licenses"])

        assert result.exit_code == 0
        assert re.search(
            # pylint: disable=line-too-long
            r"GPL-3\.0-or-later\s+GNU General Public License v3\.0 or later\s+https:\/\/spdx\.org\/licenses\/GPL-3\.0-or-later\.html\s+\n",
            result.output,
        )

    def test_json(self):
        """Output the list as JSON."""
        result = CliRunner().invoke(main, ["supported-licenses", "--json"])
        output = json.loads(result.output)

        assert result.exit_code == 0
        expected = next(
            (o for o in output if o["id"] == "GPL-3.0-or-later"), None
        )
        assert expected is not None
        assert expected["name"] == "GNU General Public License v3.0 or later"
        assert (
            expected["reference"]
            == "https://spdx.org/licenses/GPL-3.0-or-later.html"
        )



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2024 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2025 Rivos Inc.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All tests for reuse.header"""

from inspect import cleandoc

import pytest

from reuse.comment import CppCommentStyle
from reuse.copyright import CopyrightNotice, ReuseInfo, SpdxExpression
from reuse.exceptions import MissingReuseInfoError
from reuse.header import add_new_header, create_header, find_and_replace_header

# REUSE-IgnoreStart


def test_create_header_simple():
    """Create a super simple header."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert create_header(info).strip() == expected


def test_create_header_simple_with_contributor():
    """Create a super simple header."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
        contributor_lines={"John Doe"},
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileContributor: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert create_header(info).strip() == expected


def test_create_header_template_simple(template_simple):
    """Create a header with a simple template."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    expected = cleandoc(
        """
        # Hello, world!
        #
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert create_header(info, template=template_simple).strip() == expected


def test_create_header_template_no_spdx(template_no_spdx):
    """Create a header with a template that does not have all REUSE info."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )

    with pytest.raises(MissingReuseInfoError):
        create_header(info, template=template_no_spdx)


def test_create_header_template_commented(template_commented):
    """Create a header with an already-commented template."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    expected = cleandoc(
        """
        # Hello, world!
        #
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert (
        create_header(
            info,
            template=template_commented,
            template_is_commented=True,
            style=CppCommentStyle,
        ).strip()
        == expected
    )


def test_create_header_already_contains_spdx():
    """Create a new header from a header that already contains REUSE info."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    existing = cleandoc(
        """
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: MIT
        """
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        # SPDX-License-Identifier: MIT
        """
    )

    assert create_header(info, header=existing).strip() == expected


def test_create_header_existing_is_wrong():
    """If the existing header contains errors, do not raise a
    CommentCreateError.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    existing = cleandoc(
        """
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: MIT AND OR 0BSD
        """
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        # SPDX-License-Identifier: MIT AND OR 0BSD
        """
    )

    assert create_header(info, header=existing) == expected


def test_create_header_old_syntax():
    """Old copyright syntax is preserved when creating a new header."""
    info = ReuseInfo(spdx_expressions={SpdxExpression("GPL-3.0-or-later")})
    existing = cleandoc(
        """
        # Copyright John Doe
        """
    )
    expected = cleandoc(
        """
        # Copyright John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert create_header(info, header=existing).strip() == expected


def test_create_header_remove_fluff():
    """Any stuff that isn't REUSE info is removed when using create_header."""
    info = ReuseInfo(spdx_expressions={SpdxExpression("GPL-3.0-or-later")})
    existing = cleandoc(
        """
        # SPDX-FileCopyrightText: John Doe
        #
        # Hello, world!

        pass
        """
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        """
    )

    assert create_header(info, header=existing).strip() == expected


def test_add_new_header_simple():
    """Given text that already contains a header, create a new one, and preserve
    the old one.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = cleandoc(
        """
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: MIT

        pass
        """
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: MIT

        pass
        """
    )
    assert add_new_header(text, info) == expected


def test_find_and_replace_no_header():
    """Given text without header, add a header."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = "pass"
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert (
        find_and_replace_header(text, info)
        == add_new_header(text, info)
        == expected
    )


def test_find_and_replace_no_header_with_newline():
    """Given text that starts with a newline but no header, add a header."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = "\npass"
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert (
        find_and_replace_header(text, info)
        == add_new_header(text, info)
        == expected
    )


def test_find_and_replace_no_header_multiple_newlines():
    """Given text that starts with multiple newlines but no header, add a
    header."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = "\n\npass"
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later


        pass
        """
    )

    assert (
        find_and_replace_header(text, info)
        == add_new_header(text, info)
        == expected
    )


def test_find_and_replace_verbatim():
    """Replace a header with itself."""
    info = ReuseInfo()
    text = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert find_and_replace_header(text, info) == text


def test_find_and_replace_verbatim_no_newline():
    """Do not add an empty line after existing headers."""
    info = ReuseInfo()
    text = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later
        pass
        """
    )

    assert find_and_replace_header(text, info) == text


def test_find_and_replace_preserve_indentation():
    """If the first thing in a file is indented, do not change that thing's
    indentation.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = '    # Hello\n    print("world")'

    assert find_and_replace_header(text, info) == cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

            # Hello
            print("world")
        """
    )


def test_find_and_replace_newline_before_header():
    """In a scenario where the header is preceded by whitespace, remove the
    preceding whitespace.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: John Doe")
        },
    )
    text = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe

        pass
        """
    )
    text = "\n" + text
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_preserve_preceding():
    """When the SPDX header is in the middle of the file, keep it there."""
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: John Doe")
        },
    )
    text = cleandoc(
        """
        # Hello, world!

        def foo(bar):
            return bar

        # SPDX-FileCopyrightText: Jane Doe

        pass
        """
    )
    expected = cleandoc(
        """
        # Hello, world!

        def foo(bar):
            return bar

        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_keep_shebang():
    """When encountering a shebang, keep it and put the REUSE header beneath
    it.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: John Doe")
        },
    )
    text = cleandoc(
        """
        #!/usr/bin/env python3

        # SPDX-FileCopyrightText: Jane Doe

        pass
        """
    )
    expected = cleandoc(
        """
        #!/usr/bin/env python3

        # SPDX-FileCopyrightText: Jane Doe
        # SPDX-FileCopyrightText: John Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_separate_shebang():
    """When the shebang is part of the same comment as the SPDX comment,
    separate the two.
    """
    info = ReuseInfo(spdx_expressions={SpdxExpression("GPL-3.0-or-later")})
    text = cleandoc(
        """
        #!/usr/bin/env python3
        #!nix-shell -p python3
        # SPDX-FileCopyrightText: Jane Doe

        pass
        """
    )
    expected = cleandoc(
        """
        #!/usr/bin/env python3
        #!nix-shell -p python3

        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_shebang_but_no_copyright():
    """When the file contains a shebang but no copyright information, keep it at
    the top of the file.
    """
    info = ReuseInfo(spdx_expressions={SpdxExpression("GPL-3.0-or-later")})
    text = cleandoc(
        """
        #!/usr/bin/env python3

        # Hello, world!

        pass
        """
    )
    expected = cleandoc(
        """
        #!/usr/bin/env python3

        # SPDX-License-Identifier: GPL-3.0-or-later

        # Hello, world!

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_only_shebang():
    """When the file only contains a shebang, add copyright info below it."""
    info = ReuseInfo(spdx_expressions={SpdxExpression("GPL-3.0-or-later")})
    text = "#!/usr/bin/env python3"
    expected = (
        cleandoc(
            """
            #!/usr/bin/env python3

            # SPDX-License-Identifier: GPL-3.0-or-later
            """
        )
        + "\n"
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_keep_old_comment():
    """When encountering a comment that does not contain copyright and
    licensing information, preserve it below the REUSE header.
    """
    info = ReuseInfo(
        spdx_expressions={SpdxExpression("GPL-3.0-or-later")},
        copyright_notices={
            CopyrightNotice.from_string("SPDX-FileCopyrightText: Jane Doe")
        },
    )
    text = cleandoc(
        """
        # Hello, world!

        pass
        """
    )
    expected = cleandoc(
        """
        # SPDX-FileCopyrightText: Jane Doe
        #
        # SPDX-License-Identifier: GPL-3.0-or-later

        # Hello, world!

        pass
        """
    )

    assert find_and_replace_header(text, info) == expected


def test_find_and_replace_preserve_newline():
    """If the file content ends with a newline, don't remove it."""

    info = ReuseInfo()
    text = (
        cleandoc(
            """
            # SPDX-FileCopyrightText: Jane Doe
            #
            # SPDX-License-Identifier: GPL-3.0-or-later

            pass
            """
        )
        + "\n"
    )

    assert find_and_replace_header(text, info) == text


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for lint-file."""

# pylint: disable=unused-argument

from click.testing import CliRunner

from reuse.cli.main import main


class TestLintFile:
    """Tests for lint-file."""

    def test_simple(self, fake_repository):
        """A simple test to make sure it works."""
        result = CliRunner().invoke(main, ["lint-file", "src/custom.py"])
        assert result.exit_code == 0
        assert not result.output

    def test_quiet_lines_mutually_exclusive(self, empty_directory):
        """'--quiet' and '--lines' are mutually exclusive."""
        (empty_directory / "foo.py").write_text("foo")
        result = CliRunner().invoke(
            main, ["lint-file", "--quiet", "--lines", "foo"]
        )
        assert result.exit_code != 0
        assert "mutually exclusive" in result.output

    def test_no_copyright_licensing(self, fake_repository):
        """A file is correctly spotted when it has no copyright or licensing
        info.
        """
        (fake_repository / "foo.py").write_text("foo")
        result = CliRunner().invoke(main, ["lint-file", "foo.py"])
        assert result.exit_code == 1
        output = result.output
        assert "foo.py" in output
        assert "no license identifier" in output
        assert "no copyright notice" in output

    def test_path_outside_project(self, empty_directory):
        """A file can't be outside the project."""
        result = CliRunner().invoke(main, ["lint-file", ".."])
        assert result.exit_code != 0
        assert "'..' is not in" in result.output

    def test_file_not_exists(self, empty_directory):
        """A file must exist."""
        result = CliRunner().invoke(main, ["lint-file", "foo.py"])
        assert "'foo.py' does not exist" in result.output

    def test_ignored_file(self, fake_repository):
        """A corner case where a specified file is ignored. It isn't checked at
        all.
        """
        (fake_repository / "COPYING").write_text("foo")
        result = CliRunner().invoke(main, ["lint-file", "COPYING"])
        assert result.exit_code == 0

    def test_file_covered_by_toml(self, fake_repository_reuse_toml):
        """If a file is covered by REUSE.toml, use its infos."""
        (fake_repository_reuse_toml / "doc/foo.md").write_text("foo")
        result = CliRunner().invoke(main, ["lint-file", "doc/foo.md"])
        assert result.exit_code == 0



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for reuse.covered_files."""

import os
from pathlib import Path

from conftest import git, hg, pijul, posix

from reuse.covered_files import iter_files
from reuse.vcs import VCSStrategyGit, VCSStrategyHg, VCSStrategyPijul


class TestIterFiles:
    """Test the iter_files function."""

    def test_simple(self, empty_directory):
        """Given a directory with some files, yield all files."""
        (empty_directory / "foo").write_text("foo")
        (empty_directory / "bar").write_text("foo")

        assert {file_.name for file_ in iter_files(empty_directory)} == {
            "foo",
            "bar",
        }

    def test_ignore_dot_license(self, empty_directory):
        """When file and file.license are present, only yield file."""
        (empty_directory / "foo").write_text("foo")
        (empty_directory / "foo.license").write_text("foo")

        assert {file_.name for file_ in iter_files(empty_directory)} == {"foo"}

    def test_ignore_cal_license(self, empty_directory):
        """CAL licenses contain SPDX tags referencing themselves. They should be
        skipped.
        """
        (empty_directory / "CAL-1.0").write_text("foo")
        (empty_directory / "CAL-1.0.txt").write_text("foo")
        (empty_directory / "CAL-1.0-Combined-Work-Exception").write_text("foo")
        (empty_directory / "CAL-1.0-Combined-Work-Exception.txt").write_text(
            "foo"
        )

        assert not list(iter_files(empty_directory))

    def test_ignore_shl_license(self, empty_directory):
        """SHL-2.1 contains an SPDX tag referencing itself. It should be
        skipped.
        """
        (empty_directory / "SHL-2.1").write_text("foo")
        (empty_directory / "SHL-2.1.txt").write_text("foo")

        assert not list(iter_files(empty_directory))

    def test_ignore_git(self, empty_directory):
        """When the git directory is present, ignore it."""
        (empty_directory / ".git").mkdir()
        (empty_directory / ".git/config").write_text("foo")

        assert not list(iter_files(empty_directory))

    def test_ignore_hg(self, empty_directory):
        """When the hg directory is present, ignore it."""
        (empty_directory / ".hg").mkdir()
        (empty_directory / ".hg/config").write_text("foo")

        assert not list(iter_files(empty_directory))

    def test_ignore_license_copying(self, empty_directory):
        """When there are files names LICENSE, LICENSE.ext, COPYING, or
        COPYING.ext, ignore them.
        """
        (empty_directory / "LICENSE").write_text("foo")
        (empty_directory / "LICENSE.txt").write_text("foo")
        (empty_directory / "COPYING").write_text("foo")
        (empty_directory / "COPYING.txt").write_text("foo")

        assert not list(iter_files(empty_directory))

    def test_not_ignore_license_copying_no_ext(self, empty_directory):
        """Do not ignore files that start with LICENSE or COPYING and are
        followed by some non-extension text.
        """
        (empty_directory / "LICENSE_README.md").write_text("foo")
        (empty_directory / "COPYING2").write_text("foo")

        assert len(list(iter_files(empty_directory))) == 2

    @posix
    def test_ignore_symlinks(self, empty_directory):
        """All symlinks must be ignored."""
        (empty_directory / "blob").write_text("foo")
        (empty_directory / "symlink").symlink_to("blob")

        assert Path("symlink").absolute() not in iter_files(empty_directory)

    def test_ignore_zero_sized(self, empty_directory):
        """Empty files should be skipped."""
        (empty_directory / "foo").touch()

        assert Path("foo").absolute() not in iter_files(empty_directory)

    def test_include_meson_subprojects(self, empty_directory):
        """include_meson_subprojects is correctly interpreted."""
        (empty_directory / "foo.py").write_text("foo.py")
        subprojects_dir = empty_directory / "subprojects"
        subprojects_dir.mkdir()
        libfoo_dir = subprojects_dir / "libfoo"
        libfoo_dir.mkdir()
        bar_file = libfoo_dir / "bar.py"
        bar_file.write_text("pass")

        assert bar_file not in iter_files(empty_directory)
        assert bar_file in iter_files(
            empty_directory, include_meson_subprojects=True
        )

    def test_reuse_toml_ignored(self, empty_directory):
        """REUSE.toml is ignored."""
        (empty_directory / "REUSE.toml").write_text("version = 1")
        assert not list(iter_files(empty_directory))
        assert list(iter_files(empty_directory, include_reuse_tomls=True))


class TestIterFilesSubet:
    """Tests for  subset_files in iter_files."""

    def test_single(self, fake_repository):
        """Only yield the single specified file."""
        result = list(
            iter_files(
                fake_repository,
                subset_files={fake_repository / "src/custom.py"},
            )
        )
        assert result == [fake_repository / "src/custom.py"]

    def test_two(self, fake_repository):
        """Yield multiple specified files."""
        result = set(
            iter_files(
                fake_repository,
                subset_files={
                    fake_repository / "src/custom.py",
                    fake_repository / "src/exception.py",
                },
            )
        )
        assert result == {
            fake_repository / "src/custom.py",
            fake_repository / "src/exception.py",
        }

    def test_non_existent(self, fake_repository):
        """If a file does not exist, don't yield it."""
        result = list(
            iter_files(
                fake_repository,
                subset_files={
                    fake_repository / "src/custom.py",
                    fake_repository / "not_exist.py",
                    fake_repository / "also/does/not/exist.py",
                },
            )
        )
        assert result == [fake_repository / "src/custom.py"]

    def test_outside_cwd(self, fake_repository):
        """If a file is outside of the project, don't yield it."""
        result = list(
            iter_files(
                fake_repository,
                subset_files={
                    fake_repository / "src/custom.py",
                    (fake_repository / "../outside.py").resolve(),
                },
            )
        )
        assert result == [fake_repository / "src/custom.py"]

    def test_empty(self, fake_repository):
        """If no files are provided, yield nothing."""
        result = list(iter_files(fake_repository, subset_files=set()))
        assert not result

    def test_list_arg(self, fake_repository):
        """Also accepts a list argument."""
        result = list(
            iter_files(
                fake_repository,
                subset_files=[fake_repository / "src/custom.py"],
            )
        )
        assert result == [fake_repository / "src/custom.py"]

    def test_relative_path(self, fake_repository):
        """Also handles relative paths."""
        result = list(
            iter_files(fake_repository, subset_files={"src/custom.py"})
        )
        assert result == [fake_repository / "src/custom.py"]


@git
class TestAllFilesGit:
    """Test the iter_files function with git."""

    def test_simple(self, git_repository):
        """Given a Git repository where some files are ignored, do not yield
        those files.
        """
        assert Path("build/hello.py").absolute() not in iter_files(
            git_repository, vcs_strategy=VCSStrategyGit(git_repository)
        )

    def test_not_ignored_if_no_strategy(self, git_repository):
        """If no strategy is provided, the file is not ignored."""
        assert Path("build/hello.py").absolute() in iter_files(git_repository)

    def test_different_cwd(self, git_repository):
        """Given a Git repository where some files are ignored, do not yield
        those files.

        Be in a different CWD during the above.
        """
        os.chdir(git_repository / "LICENSES")
        assert Path("build/hello.py").absolute() not in iter_files(
            git_repository, vcs_strategy=VCSStrategyGit(git_repository)
        )

    def test_ignored_contains_space(self, git_repository):
        """Files that contain spaces are also ignored."""
        (git_repository / "I contain spaces.pyc").write_text("foo")
        assert Path("I contain spaces.pyc").absolute() not in iter_files(
            git_repository, vcs_strategy=VCSStrategyGit(git_repository)
        )

    @posix
    def test_ignored_contains_newline(self, git_repository):
        """Files that contain newlines are also ignored."""
        (git_repository / "hello\nworld.pyc").write_text("foo")
        assert Path("hello\nworld.pyc").absolute() not in iter_files(
            git_repository, vcs_strategy=VCSStrategyGit(git_repository)
        )

    def test_ignore_submodules(self, submodule_repository):
        """Normally ignore submodules."""
        (submodule_repository / "submodule/foo.py").write_text("foo")
        assert Path("submodule/foo.py").absolute() not in iter_files(
            submodule_repository,
            vcs_strategy=VCSStrategyGit(submodule_repository),
        )

    def test_include_submodules(self, submodule_repository):
        """If include_submodules is True, include files from the submodule."""
        (submodule_repository / "submodule/foo.py").write_text("foo")
        assert Path("submodule/foo.py").absolute() in iter_files(
            submodule_repository,
            include_submodules=True,
            vcs_strategy=VCSStrategyGit(submodule_repository),
        )

    def test_submodule_is_ignored(self, submodule_repository):
        """If a submodule is ignored, iter_files shouldn't raise an Exception"""
        (submodule_repository / "submodule/foo.py").write_text("foo")
        gitignore = submodule_repository / ".gitignore"
        contents = gitignore.read_text()
        contents += "\nsubmodule/\n"
        gitignore.write_text(contents)
        assert Path("submodule/foo.py").absolute() not in iter_files(
            submodule_repository,
            vcs_strategy=VCSStrategyGit(submodule_repository),
        )


@hg
class TestAllFilesHg:
    """Test the iter_files function with Mercurial."""

    def test_simple(self, hg_repository):
        """Given a mercurial repository where some files are ignored, do not
        yield those files.
        """
        assert Path("build/hello.py").absolute() not in iter_files(
            hg_repository, vcs_strategy=VCSStrategyHg(hg_repository)
        )

    def test_different_cwd(self, hg_repository):
        """Given a mercurial repository where some files are ignored, do not
        yield those files.

        Be in a different CWD during the above.
        """
        os.chdir(hg_repository / "LICENSES")
        assert Path("build/hello.py").absolute() not in iter_files(
            hg_repository, vcs_strategy=VCSStrategyHg(hg_repository)
        )

    def test_ignored_contains_space(self, hg_repository):
        """File names that contain spaces are also ignored."""
        (hg_repository / "I contain spaces.pyc").touch()
        assert Path("I contain spaces.pyc").absolute() not in iter_files(
            hg_repository, vcs_strategy=VCSStrategyHg(hg_repository)
        )

    @posix
    def test_ignored_contains_newline(self, hg_repository):
        """File names that contain newlines are also ignored."""
        (hg_repository / "hello\nworld.pyc").touch()
        assert Path("hello\nworld.pyc").absolute() not in iter_files(
            hg_repository, vcs_strategy=VCSStrategyHg(hg_repository)
        )


@pijul
class TestAllFilesPijul:
    """Test the iter_files function with Pijul."""

    def test_simple(self, pijul_repository):
        """Given a pijul repository where some files are ignored, do not yield
        those files.
        """
        assert Path("build/hello.py").absolute() not in iter_files(
            pijul_repository, vcs_strategy=VCSStrategyPijul(pijul_repository)
        )

    def test_iter_files_pijul_ignored_different_cwd(self, pijul_repository):
        """Given a pijul repository where some files are ignored, do not yield
        those files.

        Be in a different CWD during the above.
        """
        os.chdir(pijul_repository / "LICENSES")
        assert Path("build/hello.py").absolute() not in iter_files(
            pijul_repository, vcs_strategy=VCSStrategyPijul(pijul_repository)
        )

    def test_ignored_contains_space(self, pijul_repository):
        """File names that contain spaces are also ignored."""
        (pijul_repository / "I contain spaces.pyc").touch()
        assert Path("I contain spaces.pyc").absolute() not in iter_files(
            pijul_repository, vcs_strategy=VCSStrategyPijul(pijul_repository)
        )

    @posix
    def test_ignored_contains_newline(self, pijul_repository):
        """File names that contain newlines are also ignored."""
        (pijul_repository / "hello\nworld.pyc").touch()
        assert Path("hello\nworld.pyc").absolute() not in iter_files(
            pijul_repository, vcs_strategy=VCSStrategyPijul(pijul_repository)
        )



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2024 Skyler Grey <sky@a.starrysky.fyi>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
# SPDX-FileCopyrightText: 2024 Linnea Gräf
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for reuse.project."""

import os
import shutil
import warnings
from inspect import cleandoc
from pathlib import Path
from unittest import mock

import pytest
from conftest import RESOURCES_DIRECTORY

from reuse.copyright import (
    CopyrightNotice,
    ReuseInfo,
    SourceType,
    SpdxExpression,
)
from reuse.covered_files import iter_files
from reuse.exceptions import (
    GlobalLicensingConflictError,
    GlobalLicensingParseError,
)
from reuse.global_licensing import ReuseDep5, ReuseTOML
from reuse.project import Project

# REUSE-IgnoreStart


def test_project_not_a_directory(empty_directory):
    """Cannot create a Project without a valid directory."""
    (empty_directory / "foo.py").write_text("foo")
    with pytest.raises(NotADirectoryError):
        Project.from_directory(empty_directory / "foo.py")


def test_project_not_exists(empty_directory):
    """Cannot create a Project with a directory that doesn't exist."""
    with pytest.raises(FileNotFoundError):
        Project.from_directory(empty_directory / "foo")


def test_project_conflicting_global_licensing(empty_directory):
    """If both REUSE.toml and .reuse/dep5 exist, expect a
    GlobalLicensingConflict.
    """
    (empty_directory / "REUSE.toml").write_text("version = 1")
    (empty_directory / ".reuse").mkdir()
    shutil.copy(RESOURCES_DIRECTORY / "dep5", empty_directory / ".reuse/dep5")
    with pytest.raises(GlobalLicensingConflictError):
        Project.from_directory(empty_directory)


class TestProjectAllFiles:
    """Test Project.all_files."""

    def test_with_mock(self, monkeypatch, empty_directory):
        """Instead of testing the entire behaviour, just test that a mock is
        called and its value is returned.
        """
        mock_iter_files = mock.create_autospec(iter_files)
        mock_iter_files.return_value = "foo"
        monkeypatch.setattr("reuse.project.iter_files", mock_iter_files)

        project = Project.from_directory(empty_directory)
        result = project.all_files(empty_directory)
        assert result == "foo"

        mock_iter_files.assert_called_once_with(
            empty_directory,
            include_submodules=project.include_submodules,
            include_meson_subprojects=project.include_meson_subprojects,
            vcs_strategy=project.vcs_strategy,
        )

    def test_with_mock_implicit_dir(self, monkeypatch, empty_directory):
        """If no argument is given to Project.iter_files, project.root is
        implied.
        """
        mock_iter_files = mock.create_autospec(iter_files)
        mock_iter_files.return_value = "foo"
        monkeypatch.setattr("reuse.project.iter_files", mock_iter_files)

        project = Project.from_directory(empty_directory)
        result = project.all_files()
        assert result == "foo"

        mock_iter_files.assert_called_once_with(
            empty_directory,
            include_submodules=project.include_submodules,
            include_meson_subprojects=project.include_meson_subprojects,
            vcs_strategy=project.vcs_strategy,
        )

    def test_with_mock_includes(self, monkeypatch, empty_directory):
        """include_submodules and include_meson_subprojects are true."""
        mock_iter_files = mock.create_autospec(iter_files)
        mock_iter_files.return_value = "foo"
        monkeypatch.setattr("reuse.project.iter_files", mock_iter_files)

        project = Project.from_directory(
            empty_directory,
            include_submodules=True,
            include_meson_subprojects=True,
        )
        result = project.all_files()
        assert result == "foo"

        mock_iter_files.assert_called_once_with(
            empty_directory,
            include_submodules=project.include_submodules,
            include_meson_subprojects=project.include_meson_subprojects,
            vcs_strategy=project.vcs_strategy,
        )


def test_reuse_info_of_file_does_not_exist(fake_repository):
    """Raise FileNotFoundError when asking for the REUSE info of a file that
    does not exist.
    """
    project = Project.from_directory(fake_repository)
    with pytest.raises(FileNotFoundError):
        project.reuse_info_of(fake_repository / "does_not_exist")


def test_reuse_info_of_directory(empty_directory):
    """Raise IsADirectoryError when calling reuse_info_of on a directory."""
    (empty_directory / "src").mkdir()

    project = Project.from_directory(empty_directory)
    with pytest.raises((IsADirectoryError, PermissionError)):
        project.reuse_info_of(empty_directory / "src")


def test_reuse_info_of_unlicensed_file(fake_repository):
    """Return an empty set when asking for the REUSE information of a file that
    has no REUSE information.

    """
    (fake_repository / "foo.py").write_text("foo")
    project = Project.from_directory(fake_repository)
    assert not bool(project.reuse_info_of("foo.py"))


def test_reuse_info_of_uncommentable_file(empty_directory):
    """When a file is marked uncommentable, but does contain REUSE info, read it
    anyway.
    """
    (empty_directory / "foo.png").write_text("Copyright 2017 Jane Doe")
    project = Project.from_directory(empty_directory)
    result = project.reuse_info_of("foo.png")
    assert len(result) == 1
    assert result[0].copyright_notices


def test_reuse_info_of_only_copyright(fake_repository):
    """A file contains only a copyright line. Test whether it correctly picks
    up on that.
    """
    (fake_repository / "foo.py").write_text(
        "SPDX-FileCopyrightText: 2017 Jane Doe"
    )
    project = Project.from_directory(fake_repository)
    reuse_info = project.reuse_info_of("foo.py")[0]
    assert not any(reuse_info.spdx_expressions)
    assert len(reuse_info.copyright_notices) == 1
    assert reuse_info.copyright_notices.pop() == CopyrightNotice.from_string(
        "SPDX-FileCopyrightText: 2017 Jane Doe"
    )
    assert reuse_info.source_type == SourceType.FILE_HEADER
    assert reuse_info.source_path == "foo.py"
    assert reuse_info.path == "foo.py"


def test_reuse_info_of_toml_precedence(empty_directory):
    """When the precedence is set to toml, ignore file contents."""
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            precedence = "override"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            # The below should give a parser error. Not going to happen because
            # the file is never parsed.
            SPDX-License-Identifier: ignored AND
            SPDX-FileCopyrightText: ignored
            """
        )
    )
    project = Project.from_directory(empty_directory)
    reuse_infos = project.reuse_info_of("foo.py")
    assert len(reuse_infos) == 1
    reuse_info = reuse_infos[0]
    assert reuse_info.source_type == SourceType.REUSE_TOML
    assert reuse_info.source_path == "REUSE.toml"
    assert reuse_info.path == "foo.py"
    assert SpdxExpression("CC0-1.0") in reuse_info.spdx_expressions
    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: 2017 Jane Doe")
        in reuse_info.copyright_notices
    )


def test_reuse_info_of_closest_precedence(empty_directory):
    """When the precedence is set to closest, ignore REUSE.toml contents."""
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            precedence = "closest"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            SPDX-License-Identifier: MIT
            SPDX-FileCopyrightText: In File
            """
        )
    )
    project = Project.from_directory(empty_directory)
    reuse_infos = project.reuse_info_of("foo.py")
    assert len(reuse_infos) == 1
    reuse_info = reuse_infos[0]
    assert reuse_info.source_type == SourceType.FILE_HEADER
    assert reuse_info.source_path == "foo.py"
    assert reuse_info.path == "foo.py"
    assert SpdxExpression("MIT") in reuse_info.spdx_expressions
    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: In File")
        in reuse_info.copyright_notices
    )


def test_reuse_info_of_closest_precedence_empty(empty_directory):
    """When the precedence is set to closest, but the file is empty, use
    REUSE.toml contents.
    """
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            precedence = "closest"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").touch()
    project = Project.from_directory(empty_directory)
    reuse_infos = project.reuse_info_of("foo.py")
    assert len(reuse_infos) == 1
    reuse_info = reuse_infos[0]
    assert reuse_info.source_type == SourceType.REUSE_TOML
    assert reuse_info.source_path == "REUSE.toml"
    assert reuse_info.path == "foo.py"
    assert SpdxExpression("CC0-1.0") in reuse_info.spdx_expressions
    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: 2017 Jane Doe")
        in reuse_info.copyright_notices
    )


def test_reuse_info_of_aggregate_precedence(empty_directory):
    """When the precedence is set to aggregate, aggregate sources."""
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            precedence = "aggregate"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            SPDX-License-Identifier: MIT
            SPDX-FileCopyrightText: In File
            """
        )
    )
    project = Project.from_directory(empty_directory)
    reuse_infos = project.reuse_info_of("foo.py")
    assert len(reuse_infos) == 2
    assert reuse_infos[0].source_type != reuse_infos[1].source_type
    for reuse_info in reuse_infos:
        if reuse_info.source_type == SourceType.FILE_HEADER:
            assert reuse_info.source_type
            assert reuse_info.source_path == "foo.py"
            assert reuse_info.path == "foo.py"
            assert SpdxExpression("MIT") in reuse_info.spdx_expressions
            assert (
                CopyrightNotice.from_string("SPDX-FileCopyrightText: In File")
                in reuse_info.copyright_notices
            )
        elif reuse_info.source_type == SourceType.REUSE_TOML:
            assert reuse_info.source_path == "REUSE.toml"
            assert reuse_info.path == "foo.py"
            assert SpdxExpression("CC0-1.0") in reuse_info.spdx_expressions
            assert (
                CopyrightNotice.from_string(
                    "SPDX-FileCopyrightText: 2017 Jane Doe"
                )
                in reuse_info.copyright_notices
            )
        else:
            assert False


def test_reuse_info_of_aggregate_and_closest(empty_directory):
    """A rather tricky case. Top-level REUSE.toml says aggregate. Nearest
    REUSE.toml says closest. The top-level REUSE.toml info should now be
    aggregated with the file contents IF they exist. Else, aggregate with the
    nearest REUSE.toml info.
    """
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "src/foo.py"
            precedence = "aggregate"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "src").mkdir()
    (empty_directory / "src/REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            precedence = "closest"
            SPDX-FileCopyrightText = "2017 John Doe"
            SPDX-License-Identifier = "MIT"
            """
        )
    )
    (empty_directory / "src/foo.py").touch()
    project = Project.from_directory(empty_directory)
    assert project.reuse_info_of("src/foo.py") == [
        ReuseInfo(
            spdx_expressions={SpdxExpression("CC0-1.0")},
            copyright_notices={
                CopyrightNotice.from_string(
                    "SPDX-FileCopyrightText: 2017 Jane Doe"
                )
            },
            path="src/foo.py",
            source_path="REUSE.toml",
            source_type=SourceType.REUSE_TOML,
        ),
        ReuseInfo(
            spdx_expressions={SpdxExpression("MIT")},
            copyright_notices={
                CopyrightNotice.from_string(
                    "SPDX-FileCopyrightText: 2017 John Doe"
                )
            },
            path="src/foo.py",
            source_path="src/REUSE.toml",
            source_type=SourceType.REUSE_TOML,
        ),
    ]

    # Populate the file.
    (empty_directory / "src/foo.py").write_text(
        cleandoc(
            """
            # Copyright Example
            # SPDX-License-Identifier: 0BSD
            """
        )
    )
    assert project.reuse_info_of("src/foo.py") == [
        ReuseInfo(
            spdx_expressions={SpdxExpression("CC0-1.0")},
            copyright_notices={
                CopyrightNotice.from_string(
                    "SPDX-FileCopyrightText: 2017 Jane Doe"
                )
            },
            path="src/foo.py",
            source_path="REUSE.toml",
            source_type=SourceType.REUSE_TOML,
        ),
        ReuseInfo(
            spdx_expressions={SpdxExpression("0BSD")},
            copyright_notices={
                CopyrightNotice.from_string("Copyright Example")
            },
            path="src/foo.py",
            source_path="src/foo.py",
            source_type=SourceType.FILE_HEADER,
        ),
    ]


def test_reuse_info_of_copyright_xor_licensing(empty_directory):
    """Test a corner case where partial REUSE information is defined inside of a
    file (copyright xor licensing). Get the missing information from the
    REUSE.toml.
    """
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"

            [[annotations]]
            path = "bar.py"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            SPDX-License-Identifier: MIT
            """
        )
    )
    (empty_directory / "bar.py").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: 2017 John Doe
            """
        )
    )
    project = Project.from_directory(empty_directory)

    foo_infos = project.reuse_info_of("foo.py")
    assert len(foo_infos) == 2
    foo_toml_info = [info for info in foo_infos if info.copyright_notices][0]
    assert foo_toml_info.source_type == SourceType.REUSE_TOML
    assert not foo_toml_info.spdx_expressions
    foo_file_info = [info for info in foo_infos if info.spdx_expressions][0]
    assert foo_file_info.source_type == SourceType.FILE_HEADER
    assert not foo_file_info.copyright_notices

    bar_infos = project.reuse_info_of("bar.py")
    assert len(bar_infos) == 2
    bar_toml_info = [info for info in bar_infos if info.spdx_expressions][0]
    assert bar_toml_info.source_type == SourceType.REUSE_TOML
    assert not bar_toml_info.copyright_notices
    bar_file_info = [info for info in bar_infos if info.copyright_notices][0]
    assert bar_file_info.source_type == SourceType.FILE_HEADER
    assert not bar_file_info.spdx_expressions


def test_reuse_info_of_reuse_toml_dot_license(empty_directory):
    """Test a corner case where there is REUSE information inside of a file, its
    .license file, and REUSE.toml. Only the REUSE information from the .license
    file and REUSE.toml should be applied to this file.
    """
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "*.py"
            precedence = "aggregate"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: NONE
            """
        )
    )
    (empty_directory / "foo.py.license").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: 2017 John Doe
            """
        )
    )
    project = Project.from_directory(empty_directory)

    infos = project.reuse_info_of("foo.py")
    assert len(infos) == 2
    toml_info = [info for info in infos if info.spdx_expressions][0]
    assert toml_info.source_type == SourceType.REUSE_TOML
    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: 2017 Jane Doe")
        in toml_info.copyright_notices
    )
    assert "CC0-1.0" in str(toml_info.spdx_expressions)
    dot_license_info = [info for info in infos if not info.spdx_expressions][0]
    assert dot_license_info.source_type == SourceType.DOT_LICENSE
    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: 2017 John Doe")
        in dot_license_info.copyright_notices
    )
    assert not dot_license_info.spdx_expressions


def test_reuse_info_of_dot_license_invalid_target(empty_directory):
    """file.license is an invalid target in REUSE.toml."""
    (empty_directory / "REUSE.toml").write_text(
        cleandoc(
            """
            version = 1

            [[annotations]]
            path = "foo.py.license"
            SPDX-FileCopyrightText = "2017 Jane Doe"
            SPDX-License-Identifier = "CC0-1.0"
            """
        )
    )
    (empty_directory / "foo.py").write_text(
        cleandoc(
            """
            SPDX-FileCopyrightText: 2017 John Doe

            SPDX-License-Identifier: MIT
            """
        )
    )
    (empty_directory / "foo.py.license").write_text(
        cleandoc(
            """
            Empty
            """
        )
    )
    project = Project.from_directory(empty_directory)

    infos = project.reuse_info_of("foo.py")
    assert len(infos) == 0


def test_reuse_info_of_no_duplicates(empty_directory):
    """A file contains the same lines twice. The ReuseInfo only contains those
    lines once.
    """
    spdx_line = "SPDX-License-Identifier: GPL-3.0-or-later\n"
    copyright_line = (
        "SPDX-FileCopyrightText: 2017 Free Software Foundation Europe\n"
    )
    text = spdx_line + copyright_line

    (empty_directory / "foo.py").write_text(text * 2)
    project = Project.from_directory(empty_directory)
    reuse_info = project.reuse_info_of("foo.py")[0]
    assert len(reuse_info.spdx_expressions) == 1
    assert SpdxExpression("GPL-3.0-or-later") in reuse_info.spdx_expressions
    assert len(reuse_info.copyright_notices) == 1
    assert (
        CopyrightNotice.from_string(
            "SPDX-FileCopyrightText: 2017 Free Software Foundation Europe"
        )
        in reuse_info.copyright_notices
    )


def test_reuse_info_of_binary_succeeds(fake_repository_dep5):
    """reuse_info_of succeeds when the target is covered by dep5."""
    shutil.copy(
        RESOURCES_DIRECTORY / "fsfe.png", fake_repository_dep5 / "doc/fsfe.png"
    )

    project = Project.from_directory(fake_repository_dep5)
    reuse_info = project.reuse_info_of("doc/fsfe.png")[0]
    assert SpdxExpression("CC0-1.0") in reuse_info.spdx_expressions
    assert reuse_info.source_type == SourceType.DEP5
    assert reuse_info.path == "doc/fsfe.png"


def test_license_file_detected(empty_directory):
    """Test whether---when given a file and a license file---the license file
    is detected and read.
    """
    (empty_directory / "foo.py").write_text("foo")
    (empty_directory / "foo.py.license").write_text(
        "SPDX-FileCopyrightText: 2017 Jane Doe\nSPDX-License-Identifier: MIT\n"
    )

    project = Project.from_directory(empty_directory)
    reuse_info = project.reuse_info_of("foo.py")[0]

    assert (
        CopyrightNotice.from_string("SPDX-FileCopyrightText: 2017 Jane Doe")
        in reuse_info.copyright_notices
    )
    assert SpdxExpression("MIT") in reuse_info.spdx_expressions
    assert reuse_info.source_type == SourceType.DOT_LICENSE
    assert reuse_info.path == "foo.py"
    assert reuse_info.source_path == "foo.py.license"


def test_licenses_filename(empty_directory):
    """Detect the license identifier of a license from its stem."""
    (empty_directory / "LICENSES").mkdir()
    (empty_directory / "LICENSES/foo.txt").write_text("foo")
    project = Project.from_directory(empty_directory)
    assert "foo" in project.licenses


def test_licenses_no_extension(empty_directory):
    """Detect the license identifier of a license from its full name if it is
    in the license list.
    """
    (empty_directory / "LICENSES").mkdir()
    (empty_directory / "LICENSES/GPL-3.0-or-later").write_text("foo")
    (empty_directory / "LICENSES/MIT-3.0-or-later").write_text("foo")
    project = Project.from_directory(empty_directory)
    assert "GPL-3.0-or-later" in project.licenses
    assert "MIT-3" in project.licenses


def test_licenses_subdirectory(empty_directory):
    """Find a license in a subdirectory of LICENSES/."""
    (empty_directory / "LICENSES/sub").mkdir(parents=True)
    (empty_directory / "LICENSES/sub/MIT.txt").write_text("foo")
    project = Project.from_directory(empty_directory)
    assert "MIT" in project.licenses


def test_relative_from_root(empty_directory):
    """A simple test. Given /path/to/root/src/hello.py, return src/hello.py."""
    project = Project.from_directory(empty_directory)
    assert project.relative_from_root(project.root / "src/hello.py") == Path(
        "src/hello.py"
    )


def test_relative_from_root_no_shared_base_path(empty_directory):
    """A path can still be relative from root if the paths do not have a common
    prefix.

    For instance, if root is /path/to/root and given root/src/hello.py from the
    directory /path/to, return src/hello.py. This is a bit involved, but works
    out.
    """
    project = Project.from_directory(empty_directory)
    parent = empty_directory.parent
    os.chdir(parent)
    assert project.relative_from_root(
        Path(f"{project.root.name}/src/hello.py")
    ) == Path("src/hello.py")


def test_find_global_licensing_dep5(fake_repository_dep5):
    """Find the dep5 file. Also output a PendingDeprecationWarning."""
    with warnings.catch_warnings(record=True) as caught_warnings:
        result = Project.find_global_licensing(fake_repository_dep5)
        assert len(result) == 1
        dep5 = result[0]
        assert dep5.path == fake_repository_dep5 / ".reuse/dep5"
        assert dep5.cls == ReuseDep5

        assert len(caught_warnings) == 1
        assert issubclass(
            caught_warnings[0].category, PendingDeprecationWarning
        )


def test_find_global_licensing_reuse_toml(fake_repository_reuse_toml):
    """Find the REUSE.toml file."""
    result = Project.find_global_licensing(fake_repository_reuse_toml)
    assert len(result) == 1
    toml = result[0]
    assert toml.path == fake_repository_reuse_toml / "REUSE.toml"
    assert toml.cls == ReuseTOML


def test_find_global_licensing_reuse_toml_multiple(fake_repository_reuse_toml):
    """Find multiple REUSE.tomls."""
    (fake_repository_reuse_toml / "src/REUSE.toml").write_text("version = 1")
    result = Project.find_global_licensing(fake_repository_reuse_toml)
    assert len(result) == 2
    assert result[0].path == fake_repository_reuse_toml / "REUSE.toml"
    assert result[1].path == fake_repository_reuse_toml / "src/REUSE.toml"


def test_find_global_licensing_none(empty_directory):
    """Find no file."""
    result = Project.find_global_licensing(empty_directory)
    assert not result


def test_find_global_licensing_conflict(fake_repository_dep5):
    """Expect an error on a conflict"""
    (fake_repository_dep5 / "REUSE.toml").write_text("version = 1")
    with pytest.raises(GlobalLicensingConflictError):
        Project.find_global_licensing(fake_repository_dep5)


def test_duplicate_field_dep5(empty_directory):
    """When a duplicate field is in a dep5 file, correctly handle errors."""
    dep5_text = cleandoc(
        """
        Format: https://example.com/format/1.0
        Upstream-Name: Some project
        Upstream-Contact: Jane Doe
        Source: https://example.com/

        Files: foo.py
        Copyright: 2017 Jane Doe
        Copyright: 2017 John Doe
        License: GPL-3.0-or-later
        """
    )
    (empty_directory / ".reuse").mkdir()
    (empty_directory / ".reuse/dep5").write_text(dep5_text)

    with pytest.raises(GlobalLicensingParseError):
        Project.from_directory(empty_directory)


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2024 Nico Rikken <nico@nicorikken.eu>
# SPDX-FileCopyrightText: 2024 Sebastien Morais <github@SMoraisAnsys>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All tests for reuse.lint."""

import re
import shutil
from inspect import cleandoc

from conftest import cpython, posix

from reuse.lint import format_lines, format_plain
from reuse.project import Project
from reuse.report import ProjectReport

# REUSE-IgnoreStart


def test_lint_simple(fake_repository):
    """Extremely simple test for lint."""
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)
    assert result


def test_lint_git(git_repository):
    """Extremely simple test for lint with a git repository."""
    project = Project.from_directory(git_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)
    assert result


def test_lint_submodule(submodule_repository):
    """Extremely simple test for lint with an ignored submodule."""
    project = Project.from_directory(submodule_repository)
    (submodule_repository / "submodule/foo.c").write_text("foo")
    report = ProjectReport.generate(project)
    result = format_plain(report)
    assert result


def test_lint_submodule_included(submodule_repository):
    """Extremely simple test for lint with an included submodule."""
    project = Project.from_directory(
        submodule_repository, include_submodules=True
    )
    (submodule_repository / "submodule/foo.c").write_text("foo")
    report = ProjectReport.generate(project)
    result = format_plain(report)
    assert ":-(" in result


def test_lint_empty_directory(empty_directory):
    """An empty directory is compliant."""
    project = Project.from_directory(empty_directory)
    report = ProjectReport.generate(project)
    result = format_plain(report)
    assert result


def test_lint_deprecated(fake_repository):
    """If a repo has a deprecated license, detect it."""
    shutil.copy(
        fake_repository / "LICENSES/GPL-3.0-or-later.txt",
        fake_repository / "LICENSES/GPL-3.0.txt",
    )
    (fake_repository / "foo.py").write_text(
        "SPDX-License-Identifier: GPL-3.0\nSPDX-FileCopyrightText: Jane Doe"
    )

    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# DEPRECATED LICENSES" in result
    assert "GPL-3.0" in result
    assert "Fix deprecated licenses:" in result
    assert "spdx.org/licenses/#deprecated" in result


def test_lint_bad_license(fake_repository):
    """A bad license is detected."""
    (fake_repository / "LICENSES/foo.txt").write_text("Hello, world!")
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# BAD LICENSES" in result
    assert "foo.txt" in result
    assert "Fix bad licenses:" in result
    assert "reuse.software/faq/#custom-license" in result


def test_lint_licenses_without_extension(fake_repository):
    """A license without file extension is detected."""
    (fake_repository / "LICENSES/GPL-3.0-or-later.txt").rename(
        fake_repository / "LICENSES/GPL-3.0-or-later"
    )
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# LICENSES WITHOUT FILE EXTENSION" in result
    assert "GPL-3.0-or-later" in result
    assert "Fix licenses without file extension:" in result


def test_lint_missing_licenses(fake_repository):
    """A missing license is detected."""
    (fake_repository / "foo.py").write_text("SPDX-License-Identifier: MIT")
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# MISSING LICENSES" in result
    assert "foo.py" in result
    assert "MIT" in result
    assert "Fix missing licenses:" in result


def test_lint_unused_licenses(fake_repository):
    """An unused license is detected."""
    (fake_repository / "LICENSES/MIT.txt").write_text("foo")
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# UNUSED LICENSES" in result
    assert "Unused licenses: MIT" in result
    assert "Fix unused licenses:" in result


@cpython
@posix
def test_lint_read_errors(fake_repository):
    """A read error is detected."""
    (fake_repository / "foo.py").write_text("foo")
    (fake_repository / "foo.py").chmod(0o000)
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# READ ERRORS" in result
    assert "Could not read:" in result
    assert "foo.py" in result
    assert "Fix read errors:" in result


def test_invalid_spdx_expressions(fake_repository):
    """An invalid expression is detected."""
    (fake_repository / "foo.py").write_text(
        cleandoc(
            """
            Copyright Jane Doe

            SPDX-License-Identifier: MIT OR
            SPDX-License-Identifier: Apache-2.0 AND
            SPDX-License-Identifier: 0BSD
            """
        )
    )
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# INVALID SPDX LICENSE EXPRESSIONS" in result
    assert "foo.py' contains invalid SPDX License Expressions:" in result
    assert "Invalid SPDX License Expressions: 2" in result
    assert "Fix invalid SPDX License Expressions:" in result


def test_lint_files_without_copyright_and_licensing(fake_repository):
    """A file without copyright and licensing is detected."""
    (fake_repository / "foo.py").write_text("foo")
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_plain(report)

    assert ":-(" in result
    assert "# MISSING COPYRIGHT AND LICENSING INFORMATION" in result
    assert (
        "The following files have no copyright and licensing information:"
        in result
    )
    assert "foo.py" in result
    assert "Fix missing copyright/licensing information:" in result
    assert "reuse.software/tutorial" in result


def test_lint_json_output(fake_repository):
    """Test for lint with JSON output."""
    (fake_repository / "foo.py").write_text("SPDX-License-Identifier: MIT")
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)

    json_result = report.to_dict_lint()

    assert json_result
    # Test if all the keys are present
    assert "lint_version" in json_result
    assert "reuse_spec_version" in json_result
    assert "reuse_tool_version" in json_result
    assert "non_compliant" in json_result
    assert "files" in json_result
    assert "summary" in json_result
    assert "recommendations" in json_result
    # Test length of resulting list values
    assert len(json_result["files"]) == 9
    assert len(json_result["summary"]) == 5
    assert len(json_result["recommendations"]) == 2
    # Test result
    assert json_result["summary"]["compliant"] is False
    # Test license path
    for test_file in json_result["files"]:
        if test_file["path"] == "foo.py":
            assert test_file["spdx_expressions"][0]["value"] == "MIT"
            assert test_file["spdx_expressions"][0]["source"] == "foo.py"
            assert test_file["spdx_expressions"][0]["is_valid"]
        if test_file["path"] == "doc/usage.md":
            assert test_file["spdx_expressions"][0]["value"] == "CC0-1.0"
            assert test_file["spdx_expressions"][0]["source"] == "doc/usage.md"
            assert test_file["spdx_expressions"][0]["is_valid"]


def test_lint_lines_output(fake_repository):
    """Complete test for lint with lines output."""
    # Prepare a repository that includes all types of situations:
    # missing_licenses, unused_licenses, bad_licenses, deprecated_licenses,
    # licenses_without_extension, files_without_copyright,
    # files_without_licenses, read_errors
    (fake_repository / "invalid-license.py").write_text(
        "SPDX-License-Identifier: invalid"
    )
    (fake_repository / "no-license.py").write_text(
        "SPDX-FileCopyrightText: Jane Doe"
    )
    (fake_repository / "LICENSES" / "invalid-license-text").write_text(
        "An invalid license text"
    )
    (fake_repository / "LICENSES" / "Nokia-Qt-exception-1.1.txt").write_text(
        "Deprecated"
    )
    (fake_repository / "LICENSES" / "MIT").write_text("foo")
    (fake_repository / "file with spaces.py").write_text("foo")
    (fake_repository / "invalid-expression.py").write_text(
        cleandoc(
            """
            Copyright Jane Doe
            SPDX-License-Identifier: <invalid>"
            """
        )
    )

    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)

    lines_result = format_lines(report)
    lines_result_lines = lines_result.splitlines()

    assert len(lines_result_lines) == 12

    for line in lines_result_lines:
        assert re.match(".+: [^:]+", line)

    assert lines_result.count("invalid-license.py") == 2
    assert lines_result.count("no-license.py") == 1
    assert lines_result.count("LICENSES") == 6
    assert lines_result.count("invalid-license-text") == 3
    assert lines_result.count("Nokia-Qt-exception-1.1.txt") == 2
    assert lines_result.count("MIT") == 2
    assert lines_result.count("file with spaces.py") == 2
    assert lines_result.count("invalid-expression.py") == 2


@cpython
@posix
def test_lint_lines_read_errors(fake_repository):
    """Check read error output"""
    (fake_repository / "restricted.py").write_text("foo")
    (fake_repository / "restricted.py").chmod(0o000)
    project = Project.from_directory(fake_repository)
    report = ProjectReport.generate(project)
    result = format_lines(report)
    print(result)

    assert len(result.splitlines()) == 1
    assert "restricted.py" in result
    assert "read error" in result


# REUSE-IgnoreEnd
